@@ -81,12 +81,14 @@ from core import vm
 from frontend import consts
 from frontend import lexer
 from frontend import location
+from frontend import typed_args
 from osh import braces
 from osh import sh_expr_eval
 from osh import word_eval
+from mycpp import iolib
 from mycpp import mops
 from mycpp import mylib
-from mycpp.mylib import log, probe, switch, tagswitch
+from mycpp.mylib import log, probe, switch, tagswitch, NewDict
 from ysh import expr_eval
 from ysh import func_proc
 from ysh import val_ops
@@ -116,13 +118,15 @@ MarkLastCommands = 1 << 3
 NoDebugTrap = 1 << 4
 NoErrTrap = 1 << 5
 
+_STRICT_ERREXIT_COND_MSG = "Command conditionals should only have one status, not %s (strict_errexit, OILS-ERR-300)"
+
 
 def MakeBuiltinArgv(argv1):
     # type: (List[str]) -> cmd_value.Argv
     argv = ['']  # dummy for argv[0]
     argv.extend(argv1)
     missing = None  # type: CompoundWord
-    return cmd_value.Argv(argv, [missing] * len(argv), False, None)
+    return cmd_value.Argv(argv, [missing] * len(argv), False, None, None)
 
 
 class Deps(object):
@@ -135,7 +139,7 @@ class Deps(object):
 
 
 def _HasManyStatuses(node):
-    # type: (command_t) -> bool
+    # type: (command_t) -> Optional[command_t]
     """Code patterns that are bad for POSIX errexit.  For YSH strict_errexit.
 
     Note: strict_errexit also uses
@@ -146,7 +150,7 @@ def _HasManyStatuses(node):
         # Atoms.
         # TODO: Do we need YSH atoms here?
         if case(command_e.Simple, command_e.DBracket, command_e.DParen):
-            return False
+            return None
 
         elif case(command_e.Redirect):
             node = cast(command.Redirect, UP_node)
@@ -165,7 +169,14 @@ def _HasManyStatuses(node):
                 return _HasManyStatuses(node.children[0])
             else:
                 # Multiple parts like 'ls | wc' is disallowed
-                return True
+                return node
+
+        elif case(command_e.AndOr):
+            node = cast(command.AndOr, UP_node)
+            for c in node.children:
+                if _HasManyStatuses(c):
+                    return c
+            return None  # otherwise allow 'if true && true; ...'
 
         # - ShAssignment could be allowed, though its exit code will always be
         #   0 without command subs
@@ -173,8 +184,7 @@ def _HasManyStatuses(node):
         #   BUT could be a proc executed inside a child process, which causes a
         #   problem: the strict_errexit check has to occur at runtime and there's
         #   no way to signal it ot the parent.
-
-    return True
+    return node
 
 
 def PlusEquals(old_val, val):
@@ -263,7 +273,7 @@ class CommandEvaluator(object):
             arena,  # type: Arena
             cmd_deps,  # type: Deps
             trap_state,  # type: trap_osh.TrapState
-            signal_safe,  # type: pyos.SignalSafe
+            signal_safe,  # type: iolib.SignalSafe
     ):
         # type: (...) -> None
         """
@@ -484,16 +494,33 @@ class CommandEvaluator(object):
 
                     return result
 
-                elif redir_type == redir_arg_type_e.Here:  # here word
-                    val = self.word_ev.EvalWordToString(arg_word)
-                    assert val.tag() == value_e.Str, val
-                    # NOTE: bash and mksh both add \n
-                    result.arg = redirect_arg.HereDoc(val.s + '\n')
-                    return result
-
                 else:
                     raise AssertionError('Unknown redirect op')
 
+            elif case(redir_param_e.HereWord):
+                arg = cast(redir_param.HereWord, UP_arg)
+
+                val = self.word_ev.EvalWordToString(arg.w)
+                assert val.tag() == value_e.Str, val
+
+                assert r.op.id == Id.Redir_TLess, r.op
+                #print(arg_word)
+
+                # NOTE: bash and mksh both add \n for
+                #   read <<< 'hi'
+                #
+                # YSH doesn't do this for multi-line strings:
+                #   read <<< '''
+                #   read <<< u'''
+                #   read <<< """
+                if arg.is_multiline:
+                    s = val.s
+                else:
+                    s = val.s + '\n'
+
+                result.arg = redirect_arg.HereDoc(s)
+                return result
+
             elif case(redir_param_e.HereDoc):
                 arg = cast(redir_param.HereDoc, UP_arg)
                 w = CompoundWord(
@@ -544,11 +571,10 @@ class CommandEvaluator(object):
         if not (self.exec_opts.errexit() and self.exec_opts.strict_errexit()):
             return
 
-        if _HasManyStatuses(node):
-            node_str = ui.CommandType(node)
-            e_die(
-                "strict_errexit only allows simple commands in conditionals (got %s). "
-                % node_str, loc.Command(node))
+        bad_node = _HasManyStatuses(node)
+        if bad_node:
+            node_str = ui.CommandType(bad_node)
+            e_die(_STRICT_ERREXIT_COND_MSG % node_str, loc.Command(bad_node))
 
     def _StrictErrExitList(self, node_list):
         # type: (List[command_t]) -> None
@@ -566,12 +592,10 @@ class CommandEvaluator(object):
 
         assert len(node_list) > 0
         node = node_list[0]
-        if _HasManyStatuses(node):
-            # TODO: consolidate error message with above
-            node_str = ui.CommandType(node)
-            e_die(
-                "strict_errexit only allows simple commands in conditionals (got %s). "
-                % node_str, loc.Command(node))
+        bad_node = _HasManyStatuses(node)
+        if bad_node:
+            node_str = ui.CommandType(bad_node)
+            e_die(_STRICT_ERREXIT_COND_MSG % node_str, loc.Command(bad_node))
 
     def _EvalCondition(self, cond, blame_tok):
         # type: (condition_t, Token) -> bool
@@ -737,10 +761,15 @@ class CommandEvaluator(object):
                     with tagswitch(obj) as case:
                         if case(value_e.List):
                             obj = cast(value.List, UP_obj)
-                            index = val_ops.ToInt(lval.index,
-                                                  'List index should be Int',
-                                                  loc.Missing)
-                            obj.items[index] = rval
+                            index = expr_eval._ConvertToInt(
+                                lval.index, 'List index should be Int',
+                                loc.Missing)
+                            i = mops.BigTruncate(index)
+                            try:
+                                obj.items[i] = rval
+                            except IndexError:
+                                raise error.Expr('index out of range',
+                                                 loc.Missing)
 
                         elif case(value_e.Dict):
                             obj = cast(value.Dict, UP_obj)
@@ -810,8 +839,11 @@ class CommandEvaluator(object):
 
             if node.typed_args or node.block:  # guard to avoid allocs
                 cmd_val.proc_args = ProcArgs(node.typed_args, None, None, None)
-                func_proc.EvalTypedArgsToProc(self.expr_ev, self.mutable_opts,
-                                              node, cmd_val.proc_args)
+                func_proc.EvalTypedArgsToProc(self.expr_ev,
+                                              self.mem.CurrentFrame(),
+                                              self.mem.GlobalFrame(),
+                                              self.mutable_opts, node,
+                                              cmd_val.proc_args)
         else:
             if node.block:
                 e_die("ShAssignment builtins don't accept blocks",
@@ -827,15 +859,27 @@ class CommandEvaluator(object):
 
         # NOTE: RunSimpleCommand may never return
         if len(node.more_env):  # I think this guard is necessary?
-            is_other_special = False  # TODO: There are other special builtins too!
-            if cmd_val.tag() == cmd_value_e.Assign or is_other_special:
-                # Special builtins have their temp env persisted.
-                self._EvalTempEnv(node.more_env, 0)
-                status = self._RunSimpleCommand(cmd_val, cmd_st, run_flags)
-            else:
-                with state.ctx_Temp(self.mem):
-                    self._EvalTempEnv(node.more_env, state.SetExport)
+            if self.exec_opts.env_obj():  # YSH
+                bindings = NewDict()  # type: Dict[str, value_t]
+                with state.ctx_EnclosedFrame(self.mem, self.mem.CurrentFrame(),
+                                             self.mem.GlobalFrame(), bindings):
+                    self._EvalTempEnv(node.more_env, 0)
+
+                # Push this on the prototype chain
+                with state.ctx_EnvObj(self.mem, bindings):
                     status = self._RunSimpleCommand(cmd_val, cmd_st, run_flags)
+
+            else:  # OSH
+                is_other_special = False  # TODO: There are other special builtins too!
+                if cmd_val.tag() == cmd_value_e.Assign or is_other_special:
+                    # Special builtins have their temp env persisted.
+                    self._EvalTempEnv(node.more_env, 0)
+                    status = self._RunSimpleCommand(cmd_val, cmd_st, run_flags)
+                else:
+                    with state.ctx_Temp(self.mem):
+                        self._EvalTempEnv(node.more_env, state.SetExport)
+                        status = self._RunSimpleCommand(
+                            cmd_val, cmd_st, run_flags)
         else:
             status = self._RunSimpleCommand(cmd_val, cmd_st, run_flags)
 
@@ -1229,35 +1273,36 @@ class CommandEvaluator(object):
         status = 0  # in case we loop zero times
         with ctx_LoopLevel(self):
             while True:
-                first = it2.FirstValue()
-                #log('first %s', first)
-                if first is None:  # for StdinIterator
-                    #log('first is None')
-                    break
-
-                if first.tag() == value_e.Interrupted:
-                    self.RunPendingTraps()
-                    #log('Done running traps')
-                    continue
-
-                self.mem.SetLocalName(name1, first)
-                if name2:
-                    self.mem.SetLocalName(name2, it2.SecondValue())
-                if i_name:
-                    self.mem.SetLocalName(i_name, num.ToBig(it2.Index()))
-
-                # increment index before handling continue, etc.
-                it2.Next()
-
-                try:
-                    status = self._Execute(node.body)  # last one wins
-                except vm.IntControlFlow as e:
-                    status = 0
-                    action = e.HandleLoop()
-                    if action == flow_e.Break:
+                with state.ctx_LoopFrame(self.mem, name1.name):
+                    first = it2.FirstValue()
+                    #log('first %s', first)
+                    if first is None:  # for StdinIterator
+                        #log('first is None')
                         break
-                    elif action == flow_e.Raise:
-                        raise
+
+                    if first.tag() == value_e.Interrupted:
+                        self.RunPendingTraps()
+                        #log('Done running traps')
+                        continue
+
+                    self.mem.SetLocalName(name1, first)
+                    if name2:
+                        self.mem.SetLocalName(name2, it2.SecondValue())
+                    if i_name:
+                        self.mem.SetLocalName(i_name, num.ToBig(it2.Index()))
+
+                    # increment index before handling continue, etc.
+                    it2.Next()
+
+                    try:
+                        status = self._Execute(node.body)  # last one wins
+                    except vm.IntControlFlow as e:
+                        status = 0
+                        action = e.HandleLoop()
+                        if action == flow_e.Break:
+                            break
+                        elif action == flow_e.Raise:
+                            raise
 
         return status
 
@@ -1295,23 +1340,17 @@ class CommandEvaluator(object):
 
     def _DoShFunction(self, node):
         # type: (command.ShFunction) -> None
-        if (self.procs.Get(node.name) and
-                not self.exec_opts.redefine_proc_func()):
-            e_die(
-                "Function %s was already defined (redefine_proc_func)" %
-                node.name, node.name_tok)
+
+        # Note: shell functions can read vars from the file they're defined in
+        # But they don't appear in the module itself -- rather it is __sh_funcs__
+        # Though we could consider disallowing them though on 'import'.
         sh_func = value.Proc(node.name, node.name_tok, proc_sig.Open,
-                             node.body, None, True)
-        self.procs.SetShFunc(node.name, sh_func)
+                             node.body, None, True, self.mem.GlobalFrame())
+        self.procs.DefineShellFunc(node.name, sh_func)
 
     def _DoProc(self, node):
         # type: (Proc) -> None
         proc_name = lexer.TokenVal(node.name)
-        if (self.procs.Get(proc_name) and
-                not self.exec_opts.redefine_proc_func()):
-            e_die(
-                "Proc %s was already defined (redefine_proc_func)" % proc_name,
-                node.name)
 
         if node.sig.tag() == proc_sig_e.Closed:
             sig = cast(proc_sig.Closed, node.sig)
@@ -1321,34 +1360,20 @@ class CommandEvaluator(object):
 
         # no dynamic scope
         proc = value.Proc(proc_name, node.name, node.sig, node.body,
-                          proc_defaults, False)
-        self.procs.SetProc(proc_name, proc)
+                          proc_defaults, False, self.mem.GlobalFrame())
+        self.procs.DefineProc(proc_name, proc)
 
     def _DoFunc(self, node):
         # type: (Func) -> None
         name = lexer.TokenVal(node.name)
         lval = location.LName(name)
 
-        # Check that we haven't already defined a function
-        cell = self.mem.GetCell(name, scope_e.LocalOnly)
-        if cell and cell.val.tag() == value_e.Func:
-            if self.exec_opts.redefine_proc_func():
-                cell.readonly = False  # Ensure we can unset the value
-                did_unset = self.mem.Unset(lval, scope_e.LocalOnly)
-                assert did_unset, name
-            else:
-                e_die(
-                    "Func %s was already defined (redefine_proc_func)" % name,
-                    node.name)
-
         pos_defaults, named_defaults = func_proc.EvalFuncDefaults(
             self.expr_ev, node)
-        func_val = value.Func(name, node, pos_defaults, named_defaults, None)
+        func_val = value.Func(name, node, pos_defaults, named_defaults,
+                              self.mem.GlobalFrame())
 
-        self.mem.SetNamed(lval,
-                          func_val,
-                          scope_e.LocalOnly,
-                          flags=state.SetReadOnly)
+        self.mem.SetNamed(lval, func_val, scope_e.LocalOnly)
 
     def _DoIf(self, node):
         # type: (command.If) -> int
@@ -2091,7 +2116,7 @@ class CommandEvaluator(object):
         self.mem.SetLastStatus(status)
         return is_return, is_fatal
 
-    def EvalCommand(self, block):
+    def EvalCommandFrag(self, frag):
         # type: (command_t) -> int
         """For builtins to evaluate command args.
 
@@ -2107,17 +2132,16 @@ class CommandEvaluator(object):
 
         (Should those be more like eval 'mystring'?)
         """
-        status = 0
-        try:
-            status = self._Execute(block)  # can raise FatalRuntimeError, etc.
-        except vm.IntControlFlow as e:  # A block is more like a function.
-            # return in a block
-            if e.IsReturn():
-                status = e.StatusCode()
-            else:
-                e_die('Unexpected control flow in block', e.token)
+        return self._Execute(frag)  # can raise FatalRuntimeError, etc.
 
-        return status
+    if 0:
+
+        def EvalCommandClosure(self, cmd):
+            # type: (value.Command) -> int
+            frag = typed_args.GetCommandFrag(cmd)
+            with state.ctx_EnclosedFrame(self.mem, cmd.captured_frame,
+                                         cmd.module_frame):
+                return self.EvalCommandFrag(frag)
 
     def RunTrapsOnExit(self, mut_status):
         # type: (IntParamBox) -> None
