@@ -48,11 +48,7 @@ class _FixedWordsAction(completion.CompletionAction):
 
 
 class _DynamicProcDictAction(completion.CompletionAction):
-    """For completing from proc and aliases dicts, which are mutable.
-
-    Note: this is the same as _FixedWordsAction now, but won't be when the code
-    is statically typed!
-    """
+    """For completing shell functions/procs/invokables."""
 
     def __init__(self, d):
         # type: (state.Procs) -> None
@@ -60,7 +56,7 @@ class _DynamicProcDictAction(completion.CompletionAction):
 
     def Matches(self, comp):
         # type: (Api) -> Iterator[str]
-        for name in self.d.GetNames():
+        for name in self.d.InvokableNames():
             if name.startswith(comp.to_complete):
                 yield name
 
@@ -70,11 +66,7 @@ class _DynamicProcDictAction(completion.CompletionAction):
 
 
 class _DynamicStrDictAction(completion.CompletionAction):
-    """For completing from proc and aliases dicts, which are mutable.
-
-    Note: this is the same as _FixedWordsAction now, but won't be when the code
-    is statically typed!
-    """
+    """For completing from the alias dicts, which is mutable."""
 
     def __init__(self, d):
         # type: (Dict[str, str]) -> None
@@ -139,9 +131,11 @@ class SpecBuilder(object):
         # obviously it's better to check here.
         if arg.F is not None:
             func_name = arg.F
-            func = cmd_ev.procs.Get(func_name)
+            func = cmd_ev.procs.GetShellFunc(func_name)
             if func is None:
-                raise error.Usage('function %r not found' % func_name,
+                # Note: we will have a different protocol for YSH procs and invokables
+                # The ideal thing would be some kind of generator ...
+                raise error.Usage('shell function %r not found' % func_name,
                                   loc.Missing)
             actions.append(
                 completion.ShellFuncAction(cmd_ev, func, self.comp_lookup))
