@@ -14,6 +14,7 @@ from _devbuild.gen.syntax_asdl import (
     loc,
     loc_t,
 )
+from _devbuild.gen.value_asdl import value, value_e
 from builtin import hay_ysh
 from core import dev
 from core import error
@@ -22,13 +23,17 @@ from core.error import e_die, e_die_status
 from core import pyos
 from core import pyutil
 from core import state
-from display import ui
 from core import vm
+from display import ui
 from frontend import consts
 from frontend import lexer
-from mycpp.mylib import log, print_stderr
+from mycpp import mylib
+from mycpp.mylib import log, print_stderr, tagswitch
+from pylib import os_path
+from pylib import path_stat
 
 import posix_ as posix
+from posix_ import X_OK  # translated directly to C macro
 
 from typing import cast, Dict, List, Tuple, Optional, TYPE_CHECKING
 if TYPE_CHECKING:
@@ -38,11 +43,116 @@ if TYPE_CHECKING:
     from builtin import trap_osh
     from core import optview
     from core import state
-    from core.vm import _Builtin
 
 _ = log
 
 
+def LookupExecutable(name, path_dirs, exec_required=True):
+    # type: (str, List[str], bool) -> Optional[str]
+    """
+    Returns either
+    - the name if it's a relative path that exists
+    - the executable name resolved against path_dirs
+    - None if not found
+    """
+    if len(name) == 0:  # special case for "$(true)"
+        return None
+
+    if '/' in name:
+        return name if path_stat.exists(name) else None
+
+    for path_dir in path_dirs:
+        full_path = os_path.join(path_dir, name)
+        if exec_required:
+            found = posix.access(full_path, X_OK)
+        else:
+            found = path_stat.exists(full_path)
+
+        if found:
+            return full_path
+
+    return None
+
+
+class SearchPath(object):
+    """For looking up files in $PATH or ENV.PATH"""
+
+    def __init__(self, mem, exec_opts):
+        # type: (state.Mem, optview.Exec) -> None
+        self.mem = mem
+        # TODO: remove exec_opts
+        self.cache = {}  # type: Dict[str, str]
+
+    def _GetPath(self):
+        # type: () -> List[str]
+
+        # In YSH, we read from ENV.PATH
+        s = self.mem.env_config.Get('PATH')
+        if s is None:
+            return []  # treat as empty path
+
+        # TODO: Could cache this to avoid split() allocating all the time.
+        return s.split(':')
+
+    def LookupOne(self, name, exec_required=True):
+        # type: (str, bool) -> Optional[str]
+        """
+        Returns the path itself (if relative path), the resolved path, or None.
+        """
+        return LookupExecutable(name,
+                                self._GetPath(),
+                                exec_required=exec_required)
+
+    def LookupReflect(self, name, do_all):
+        # type: (str, bool) -> List[str]
+        """
+        Like LookupOne(), with an option for 'type -a' to return all paths.
+        """
+        if len(name) == 0:  # special case for "$(true)"
+            return []
+
+        if '/' in name:
+            if path_stat.exists(name):
+                return [name]
+            else:
+                return []
+
+        results = []  # type: List[str]
+        for path_dir in self._GetPath():
+            full_path = os_path.join(path_dir, name)
+            if path_stat.exists(full_path):
+                results.append(full_path)
+                if not do_all:
+                    return results
+
+        return results
+
+    def CachedLookup(self, name):
+        # type: (str) -> Optional[str]
+        #log('name %r', name)
+        if name in self.cache:
+            return self.cache[name]
+
+        full_path = self.LookupOne(name)
+        if full_path is not None:
+            self.cache[name] = full_path
+        return full_path
+
+    def MaybeRemoveEntry(self, name):
+        # type: (str) -> None
+        """When the file system changes."""
+        mylib.dict_erase(self.cache, name)
+
+    def ClearCache(self):
+        # type: () -> None
+        """For hash -r."""
+        self.cache.clear()
+
+    def CachedCommands(self):
+        # type: () -> List[str]
+        return self.cache.values()
+
+
 class _ProcessSubFrame(object):
     """To keep track of diff <(cat 1) <(cat 2) > >(tac)"""
 
@@ -111,8 +221,8 @@ class ShellExecutor(vm._Executor):
             mutable_opts,  # type: state.MutableOpts
             procs,  # type: state.Procs
             hay_state,  # type: hay_ysh.HayState
-            builtins,  # type: Dict[int, _Builtin]
-            search_path,  # type: state.SearchPath
+            builtins,  # type: Dict[int, vm._Builtin]
+            search_path,  # type: SearchPath
             ext_prog,  # type: process.ExternalProgram
             waiter,  # type: process.Waiter
             tracer,  # type: dev.Tracer
@@ -199,14 +309,19 @@ class ShellExecutor(vm._Executor):
         """
         self.tracer.OnBuiltin(builtin_id, cmd_val.argv)
 
-        builtin_func = self.builtins[builtin_id]
+        builtin_proc = self.builtins[builtin_id]
+
+        return self.RunBuiltinProc(builtin_proc, cmd_val)
+
+    def RunBuiltinProc(self, builtin_proc, cmd_val):
+        # type: (vm._Builtin, cmd_value.Argv) -> int
 
         io_errors = []  # type: List[error.IOError_OSError]
         with vm.ctx_FlushStdout(io_errors):
             # note: could be second word, like 'builtin read'
             with ui.ctx_Location(self.errfmt, cmd_val.arg_locs[0]):
                 try:
-                    status = builtin_func.Run(cmd_val)
+                    status = builtin_proc.Run(cmd_val)
                     assert isinstance(status, int)
                 except (IOError, OSError) as e:
                     self.errfmt.PrintMessage(
@@ -271,16 +386,13 @@ class ShellExecutor(vm._Executor):
             #  e_die_status(status, 'special builtin failed')
             return status
 
-        call_procs = not (run_flags & NO_CALL_PROCS)
         # Builtins like 'true' can be redefined as functions.
+        call_procs = not (run_flags & NO_CALL_PROCS)
         if call_procs:
-            # TODO: Look shell functions in self.sh_funcs, but procs are
-            # value.Proc in the var namespace.
-            # Pitfall: What happens if there are two of the same name?  I guess
-            # that's why you have = and 'type' inspect them
+            proc_val, self_obj = self.procs.GetInvokable(arg0)
+            cmd_val.self_obj = self_obj  # MAYBE bind self
 
-            proc_node = self.procs.Get(arg0)
-            if proc_node is not None:
+            if proc_val is not None:
                 if self.exec_opts.strict_errexit():
                     disabled_tok = self.mutable_opts.ErrExitDisabledToken()
                     if disabled_tok:
@@ -293,9 +405,25 @@ class ShellExecutor(vm._Executor):
                             "Use 'try' or wrap it in a process with $0 myproc",
                             arg0_loc)
 
-                with dev.ctx_Tracer(self.tracer, 'proc', argv):
-                    # NOTE: Functions could call 'exit 42' directly, etc.
-                    status = self.cmd_ev.RunProc(proc_node, cmd_val)
+                with tagswitch(proc_val) as case:
+                    if case(value_e.BuiltinProc):
+                        # Handle the special case of the BUILTIN proc
+                        # module_ysh.ModuleInvoke, which is returned on the Obj
+                        # created by 'use util.ysh'
+                        builtin_proc = cast(value.BuiltinProc, proc_val)
+                        b = cast(vm._Builtin, builtin_proc.builtin)
+                        status = self.RunBuiltinProc(b, cmd_val)
+
+                    elif case(value_e.Proc):
+                        proc = cast(value.Proc, proc_val)
+                        with dev.ctx_Tracer(self.tracer, 'proc', argv):
+                            # NOTE: Functions could call 'exit 42' directly, etc.
+                            status = self.cmd_ev.RunProc(proc, cmd_val)
+
+                    else:
+                        # GetInvokable() should only return 1 of 2 things
+                        raise AssertionError()
+
                 return status
 
         # Notes:
@@ -324,7 +452,7 @@ class ShellExecutor(vm._Executor):
             cmd_st.show_code = True  # this is a "leaf" for errors
             return self.RunBuiltin(builtin_id, cmd_val)
 
-        environ = self.mem.GetExported()  # Include temporary variables
+        environ = self.mem.GetEnv()  # Include temporary variables
 
         if cmd_val.proc_args:
             e_die(
@@ -333,7 +461,7 @@ class ShellExecutor(vm._Executor):
 
         # Resolve argv[0] BEFORE forking.
         if run_flags & USE_DEFAULT_PATH:
-            argv0_path = state.LookupExecutable(arg0, DEFAULT_PATH)
+            argv0_path = LookupExecutable(arg0, DEFAULT_PATH)
         else:
             argv0_path = self.search_path.CachedLookup(arg0)
         if argv0_path is None:
