@@ -46,7 +46,7 @@ from _devbuild.gen.runtime_asdl import (
 )
 from _devbuild.gen.value_asdl import (value, value_e, value_t, y_lvalue,
                                       y_lvalue_e, y_lvalue_t, IntBox, LeftName,
-                                      Obj)
+                                      Obj, cmd_frag)
 from core import error
 from core.error import e_die, e_die_status
 from core import num
@@ -97,9 +97,12 @@ def _ConvertToInt(val, msg, blame_loc):
 
         elif case(value_e.Str):
             val = cast(value.Str, UP_val)
-            if match.LooksLikeInteger(val.s):
-                # TODO: Handle ValueError
-                return mops.FromStr(val.s)
+            if match.LooksLikeYshInt(val.s):
+                s = val.s.replace('_', '')
+                ok, i = mops.FromStr2(s)
+                if not ok:
+                    e_die("Integer too big: %s" % s, blame_loc)
+                return i
 
     raise error.TypeErr(val, msg, blame_loc)
 
@@ -118,12 +121,17 @@ def _ConvertToNumber(val):
 
         elif case(value_e.Str):
             val = cast(value.Str, UP_val)
-            if match.LooksLikeInteger(val.s):
-                # TODO: Handle ValueError
-                return coerced_e.Int, mops.FromStr(val.s), -1.0
 
-            if match.LooksLikeFloat(val.s):
-                return coerced_e.Float, mops.MINUS_ONE, float(val.s)
+            if match.LooksLikeYshInt(val.s):
+                s = val.s.replace('_', '')
+                ok, i = mops.FromStr2(s)
+                if not ok:
+                    e_die("Integer too big: %s" % s, loc.Missing)
+                return coerced_e.Int, i, -1.0
+
+            if match.LooksLikeYshFloat(val.s):
+                s = val.s.replace('_', '')
+                return coerced_e.Float, mops.MINUS_ONE, float(s)
 
     return coerced_e.Neither, mops.MINUS_ONE, -1.0
 
@@ -198,12 +206,10 @@ class ExprEvaluator(object):
 
     def EvalAugmented(self, lval, rhs_val, op, which_scopes):
         # type: (y_lvalue_t, value_t, Token, scope_t) -> None
-        """ setvar x +=1, setvar L[0] -= 1 
+        """ setvar x += 1, setvar L[0] -= 1 
 
         Called by CommandEvaluator
         """
-        # TODO: It might be nice to do auto d[x] += 1 too
-
         UP_lval = lval
         with tagswitch(lval) as case:
             if case(y_lvalue_e.Local):  # setvar x += 1
@@ -228,9 +234,11 @@ class ExprEvaluator(object):
                 with tagswitch(obj) as case:
                     if case(value_e.List):
                         obj = cast(value.List, UP_obj)
-                        index = val_ops.ToInt(lval.index,
-                                              'List index should be Int',
-                                              loc.Missing)
+                        i1 = _ConvertToInt(lval.index,
+                                           'List index should be Int',
+                                           loc.Missing)
+                        # TODO: don't truncate
+                        index = mops.BigTruncate(i1)
                         try:
                             lhs_val_ = obj.items[index]
                         except IndexError:
@@ -320,7 +328,7 @@ class ExprEvaluator(object):
                 obj = self._EvalLeftLocalOrGlobal(lhs.obj, which_scopes)
                 index = self._EvalExpr(lhs.index)
 
-                return self._EvalSubscript(obj, index)
+                return self._EvalSubscript(obj, index, lhs.left)
 
             elif case(expr_e.Attribute):
                 lhs = cast(Attribute, UP_lhs)
@@ -367,6 +375,19 @@ class ExprEvaluator(object):
             else:
                 raise AssertionError()
 
+    def EvalExprClosure(self, expr_val, blame_loc):
+        # type: (value.Expr, loc_t) -> value_t
+        """
+        Used by user-facing APIs that take value.Expr closures:
+
+        var i = 42
+        var x = io->evalExpr(^[i + 1])
+        var x = s.replace(pat, ^"- $0 $i -")
+        """
+        with state.ctx_EnclosedFrame(self.mem, expr_val.captured_frame,
+                                     expr_val.module_frame, None):
+            return self.EvalExpr(expr_val.e, blame_loc)
+
     def EvalExpr(self, node, blame_loc):
         # type: (expr_t, loc_t) -> value_t
         """Public API for _EvalExpr to ensure command_sub_errexit"""
@@ -390,13 +411,12 @@ class ExprEvaluator(object):
 
         with switch(part.left.id) as case:
             if case(Id.Left_DollarBracket):  # $[join(x)]
-                s = val_ops.Stringify(val, loc.WordPart(part))
+                s = val_ops.Stringify(val, loc.WordPart(part), 'Expr sub ')
                 return Piece(s, False, False)
 
             elif case(Id.Lit_AtLBracket):  # @[split(x)]
-                strs = val_ops.ToShellArray(val,
-                                            loc.WordPart(part),
-                                            prefix='Expr splice ')
+                strs = val_ops.ToShellArray(val, loc.WordPart(part),
+                                            'Expr splice ')
                 return part_value.Array(strs)
 
             else:
@@ -452,6 +472,16 @@ class ExprEvaluator(object):
 
         return val
 
+    def _CallMetaMethod(self, func_val, pos_args, blame_loc):
+        # type: (value_t, List[value_t], loc_t) -> value_t
+
+        named_args = {}  # type: Dict[str, value_t]
+        arg_list = ArgList.CreateNull()  # There's no call site
+        rd = typed_args.Reader(pos_args, named_args, None, arg_list)
+        rd.SetFallbackLocation(blame_loc)
+        # errors propagate
+        return self._CallFunc(func_val, rd)
+
     def SpliceValue(self, val, part):
         # type: (value_t, word_part.Splice) -> List[str]
         """ write -- @myvar """
@@ -721,17 +751,13 @@ class ExprEvaluator(object):
                 result = not val_ops.Contains(left, right)
 
             elif op.id == Id.Expr_Is:
-                if left.tag() != right.tag():
-                    raise error.TypeErrVerbose('Mismatched types', op)
                 result = left is right
 
             elif op.id == Id.Node_IsNot:
-                if left.tag() != right.tag():
-                    raise error.TypeErrVerbose('Mismatched types', op)
                 result = left is not right
 
             elif op.id == Id.Expr_DTilde:
-                # no extglob in Oil language; use eggex
+                # no extglob in YSH; use eggex
                 if left.tag() != value_e.Str:
                     raise error.TypeErrVerbose('LHS must be Str', op)
 
@@ -788,10 +814,17 @@ class ExprEvaluator(object):
 
                     elif case(value_e.Int):
                         right = cast(value.Int, UP_right)
-                        if not left2.isdigit():
+
+                        # Note: this logic is similar to _ConvertToInt(left2)
+                        if not match.LooksLikeYshInt(left2):
                             return value.Bool(False)
 
-                        eq = mops.Equal(mops.FromStr(left2), right.i)
+                        left2 = left2.replace('_', '')
+                        ok, left_i = mops.FromStr2(left2)
+                        if not ok:
+                            e_die('Integer too big: %s' % left2, op)
+
+                        eq = mops.Equal(left_i, right.i)
                         return value.Bool(eq)
 
                 e_die('~== expects Str, Int, or Bool on the right', op)
@@ -861,7 +894,7 @@ class ExprEvaluator(object):
                 to_call = func.func
                 pos_args, named_args = func_proc._EvalArgList(self,
                                                               node.args,
-                                                              me=func.me)
+                                                              self_val=func.me)
                 rd = typed_args.Reader(pos_args,
                                        named_args,
                                        None,
@@ -873,8 +906,8 @@ class ExprEvaluator(object):
 
         return self._CallFunc(to_call, rd)
 
-    def _EvalSubscript(self, obj, index):
-        # type: (value_t, value_t) -> value_t
+    def _EvalSubscript(self, obj, index, blame_loc):
+        # type: (value_t, value_t, loc_t) -> value_t
 
         UP_obj = obj
         UP_index = index
@@ -898,45 +931,52 @@ class ExprEvaluator(object):
                         try:
                             return value.Str(obj.s[i])
                         except IndexError:
-                            # TODO: expr.Subscript has no error location
-                            raise error.Expr('index out of range', loc.Missing)
+                            raise error.Expr('index out of range', blame_loc)
 
                     else:
                         raise error.TypeErr(index,
                                             'Str index expected Int or Slice',
-                                            loc.Missing)
+                                            blame_loc)
 
             elif case(value_e.List):
                 obj = cast(value.List, UP_obj)
+
+                big_i = mops.ZERO
                 with tagswitch(index) as case2:
                     if case2(value_e.Slice):
                         index = cast(value.Slice, UP_index)
 
-                        lower = index.lower.i if index.lower else 0
-                        upper = index.upper.i if index.upper else len(
-                            obj.items)
+                        lower = (index.lower.i if index.lower else 0)
+                        upper = (index.upper.i
+                                 if index.upper else len(obj.items))
                         return value.List(obj.items[lower:upper])
 
                     elif case2(value_e.Int):
                         index = cast(value.Int, UP_index)
-                        i = mops.BigTruncate(index.i)
-                        try:
-                            return obj.items[i]
-                        except IndexError:
-                            # TODO: expr.Subscript has no error location
-                            raise error.Expr('List index out of range: %d' % i,
-                                             loc.Missing)
+                        big_i = index.i
+
+                    elif case2(value_e.Str):
+                        index = cast(value.Str, UP_index)
+                        big_i = _ConvertToInt(index, 'List index expected Int',
+                                              blame_loc)
 
                     else:
                         raise error.TypeErr(
-                            index, 'List index expected Int or Slice',
-                            loc.Missing)
+                            index, 'List index expected Int, Str, or Slice',
+                            blame_loc)
+
+                i = mops.BigTruncate(big_i)  # TODO: don't truncate
+                try:
+                    return obj.items[i]
+                except IndexError:
+                    raise error.Expr('List index out of range: %d' % i,
+                                     blame_loc)
 
             elif case(value_e.Dict):
                 obj = cast(value.Dict, UP_obj)
                 if index.tag() != value_e.Str:
                     raise error.TypeErr(index, 'Dict index expected Str',
-                                        loc.Missing)
+                                        blame_loc)
 
                 index = cast(value.Str, UP_index)
                 try:
@@ -944,10 +984,20 @@ class ExprEvaluator(object):
                 except KeyError:
                     # TODO: expr.Subscript has no error location
                     raise error.Expr('Dict entry not found: %r' % index.s,
-                                     loc.Missing)
+                                     blame_loc)
 
-        raise error.TypeErr(obj, 'Subscript expected Str, List, or Dict',
-                            loc.Missing)
+            elif case(value_e.Obj):
+                obj = cast(Obj, UP_obj)
+
+                index_method = val_ops.IndexMetaMethod(obj)
+                if index_method is not None:
+                    pos_args = [obj, index]
+                    return self._CallMetaMethod(index_method, pos_args,
+                                                blame_loc)
+
+        raise error.TypeErr(
+            obj, 'Subscript expected one of (Str List Dict, indexable Obj)',
+            blame_loc)
 
     def _ChainedLookup(self, obj, current, attr_name):
         # type: (Obj, Obj, str) -> Optional[value_t]
@@ -1046,8 +1096,8 @@ class ExprEvaluator(object):
                 # - found in the properties, not in the prototype chain (not
                 #   sure if this error is common.)
                 raise error.Expr(
-                    "Mutating method %r not found on Obj prototype chain" % mut_name,
-                    node.attr)
+                    "Mutating method %r not found on Obj prototype chain" %
+                    mut_name, node.attr)
             else:
                 # Look up methods on builtin types
                 # TODO: These should also be called M/append, M/erase, etc.
@@ -1134,7 +1184,7 @@ class ExprEvaluator(object):
 
             elif case(expr_e.Place):
                 node = cast(expr.Place, UP_node)
-                frame = self.mem.TopNamespace()
+                frame = self.mem.CurrentFrame()
                 return value.Place(LeftName(node.var_name, node.blame_tok),
                                    frame)
 
@@ -1143,8 +1193,10 @@ class ExprEvaluator(object):
 
                 id_ = node.left_token.id
                 if id_ == Id.Left_CaretParen:  # ^(echo block literal)
-                    # TODO: Propgate location info?
-                    return value.Command(node.child)
+                    # TODO: Propagate location info with ^(
+                    return value.Command(cmd_frag.Expr(node.child),
+                                         self.mem.CurrentFrame(),
+                                         self.mem.GlobalFrame())
                 else:
                     stdout_str = self.shell_ex.RunCommandSub(node)
                     if id_ == Id.Left_AtParen:  # @(seq 3)
@@ -1216,16 +1268,16 @@ class ExprEvaluator(object):
                 upper = None  # type: Optional[IntBox]
 
                 if node.lower:
-                    msg = 'Slice begin should be Int'
-                    i = val_ops.ToInt(self._EvalExpr(node.lower), msg,
-                                      loc.Missing)
-                    lower = IntBox(i)
+                    i1 = _ConvertToInt(self._EvalExpr(node.lower),
+                                       'Slice begin should be Int', node.op)
+                    # TODO: don't truncate
+                    lower = IntBox(mops.BigTruncate(i1))
 
                 if node.upper:
-                    msg = 'Slice end should be Int'
-                    i = val_ops.ToInt(self._EvalExpr(node.upper), msg,
-                                      loc.Missing)
-                    upper = IntBox(i)
+                    i1 = _ConvertToInt(self._EvalExpr(node.upper),
+                                       'Slice end should be Int', node.op)
+                    # TODO: don't truncate
+                    upper = IntBox(mops.BigTruncate(i1))
 
                 return value.Slice(lower, upper)
 
@@ -1235,13 +1287,17 @@ class ExprEvaluator(object):
                 assert node.lower is not None
                 assert node.upper is not None
 
-                msg = 'Range begin should be Int'
-                i = val_ops.ToInt(self._EvalExpr(node.lower), msg, loc.Missing)
+                i1 = _ConvertToInt(self._EvalExpr(node.lower),
+                                   'Range begin should be Int', node.op)
 
-                msg = 'Range end should be Int'
-                j = val_ops.ToInt(self._EvalExpr(node.upper), msg, loc.Missing)
+                i2 = _ConvertToInt(self._EvalExpr(node.upper),
+                                   'Range end should be Int', node.op)
 
-                return value.Range(i, j)
+                if node.op.id == Id.Expr_DDotEqual:  # Closed range
+                    i2 = mops.Add(i2, mops.ONE)
+
+                # TODO: Don't truncate
+                return value.Range(mops.BigTruncate(i1), mops.BigTruncate(i2))
 
             elif case(expr_e.Compare):
                 node = cast(expr.Compare, UP_node)
@@ -1301,7 +1357,8 @@ class ExprEvaluator(object):
 
             elif case(expr_e.Literal):  # ^[1 + 2]
                 node = cast(expr.Literal, UP_node)
-                return value.Expr(node.inner)
+                return value.Expr(node.inner, self.mem.CurrentFrame(),
+                                  self.mem.GlobalFrame())
 
             elif case(expr_e.Lambda):  # |x| x+1 syntax is reserved
                 # TODO: Location information for |, or func
@@ -1316,7 +1373,7 @@ class ExprEvaluator(object):
                 node = cast(Subscript, UP_node)
                 obj = self._EvalExpr(node.obj)
                 index = self._EvalExpr(node.index)
-                return self._EvalSubscript(obj, index)
+                return self._EvalSubscript(obj, index, node.left)
 
             elif case(expr_e.Attribute):  # obj->method or mydict.key
                 node = cast(Attribute, UP_node)
