@@ -4,12 +4,10 @@ func_misc.py
 """
 from __future__ import print_function
 
-from _devbuild.gen.runtime_asdl import (scope_e)
 from _devbuild.gen.value_asdl import (value, value_e, value_t, value_str, Obj)
 
 from core import error
 from core import num
-from core import state
 from display import pp_value
 from display import ui
 from core import vm
@@ -19,7 +17,6 @@ from frontend import typed_args
 from mycpp import mops
 from mycpp import mylib
 from mycpp.mylib import NewDict, iteritems, log, tagswitch
-from ysh import expr_eval
 from ysh import val_ops
 
 from typing import TYPE_CHECKING, Dict, List, Optional, cast
@@ -31,7 +28,7 @@ _ = log
 
 
 class Object(vm._Callable):
-    """Create a value.Obj
+    """OLD API to a value.Obj
 
     The order of params follows JavaScript's Object.create():
         var obj = Object(prototype, props)
@@ -45,6 +42,8 @@ class Object(vm._Callable):
         # type: (typed_args.Reader) -> value_t
 
         prototype = rd.PosValue()
+        proto_loc = rd.BlamePos()
+
         props = rd.PosDict()
         rd.Done()
 
@@ -58,7 +57,49 @@ class Object(vm._Callable):
                 chain = prototype
             else:
                 raise error.TypeErr(prototype, 'Object() expected Obj or Null',
-                                    rd.BlamePos())
+                                    proto_loc)
+
+        return Obj(chain, props)
+
+
+class Obj_call(vm._Callable):
+    """New API to create a value.Obj
+
+    It has a more natural order
+        var obj = Obj(props, prototype)
+
+    Until we have __call__, it's Obj:
+        var obj = Obj.new(props, prototype)
+    """
+
+    def __init__(self):
+        # type: () -> None
+        pass
+
+    def Call(self, rd):
+        # type: (typed_args.Reader) -> value_t
+
+        props = rd.PosDict()
+
+        prototype = rd.OptionalValue()
+        proto_loc = rd.BlamePos()
+
+        rd.Done()
+
+        chain = None  # type: Optional[Obj]
+
+        if prototype is not None:
+            UP_prototype = prototype
+            with tagswitch(prototype) as case:
+                if case(value_e.Null):  # Obj({}, null)
+                    pass
+                elif case(value_e.Obj):
+                    prototype = cast(Obj, UP_prototype)
+                    chain = prototype
+                else:
+                    raise error.TypeErr(prototype,
+                                        'Object() expected Obj or Null',
+                                        proto_loc)
 
         return Obj(chain, props)
 
@@ -138,6 +179,11 @@ class Type(vm._Callable):
         val = rd.PosValue()
         rd.Done()
 
+        # TODO: assert it's not Undef, Interrupted, Slice
+        # Then return an Obj type
+        #
+        # It would be nice if they were immutable, if we didn't have to create
+        # 23-24 dicts and 23-24 Obj on startup?
         return value.Str(ui.ValType(val))
 
 
@@ -157,7 +203,7 @@ class Join(vm._Callable):
 
         strs = []  # type: List[str]
         for i, el in enumerate(li):
-            strs.append(val_ops.Stringify(el, rd.LeftParenToken()))
+            strs.append(val_ops.Stringify(el, rd.LeftParenToken(), 'join() '))
 
         return value.Str(delim.join(strs))
 
@@ -234,11 +280,17 @@ class Int(vm._Callable):
 
             elif case(value_e.Str):
                 val = cast(value.Str, UP_val)
-                if not match.LooksLikeInteger(val.s):
+                if not match.LooksLikeYshInt(val.s):
                     raise error.Expr("Can't convert %s to Int" % val.s,
                                      rd.BlamePos())
 
-                return value.Int(mops.FromStr(val.s))
+                s = val.s.replace('_', '')
+                ok, big_int = mops.FromStr2(s)
+                if not ok:
+                    raise error.Expr("Integer too big: %s" % val.s,
+                                     rd.BlamePos())
+
+                return value.Int(big_int)
 
         raise error.TypeErr(val, 'int() expected Bool, Int, Float, or Str',
                             rd.BlamePos())
@@ -267,7 +319,7 @@ class Float(vm._Callable):
 
             elif case(value_e.Str):
                 val = cast(value.Str, UP_val)
-                if not match.LooksLikeFloat(val.s):
+                if not match.LooksLikeYshFloat(val.s):
                     raise error.Expr('Cannot convert %s to Float' % val.s,
                                      rd.BlamePos())
 
@@ -289,23 +341,13 @@ class Str_(vm._Callable):
         val = rd.PosValue()
         rd.Done()
 
-        # TODO: Should we call Stringify here?  That would handle Eggex.
-
-        UP_val = val
         with tagswitch(val) as case:
-            if case(value_e.Int):
-                val = cast(value.Int, UP_val)
-                return value.Str(mops.ToStr(val.i))
-
-            elif case(value_e.Float):
-                val = cast(value.Float, UP_val)
-                return value.Str(str(val.f))
-
-            elif case(value_e.Str):
+            # Avoid extra allocation
+            if case(value_e.Str):
                 return val
-
-        raise error.TypeErr(val, 'str() expected Str, Int, or Float',
-                            rd.BlamePos())
+            else:
+                s = val_ops.Stringify(val, rd.LeftParenToken(), 'str() ')
+                return value.Str(s)
 
 
 class List_(vm._Callable):
@@ -367,29 +409,37 @@ class DictFunc(vm._Callable):
         UP_val = val
         with tagswitch(val) as case:
             if case(value_e.Dict):
-                d = NewDict()  # type: Dict[str, value_t]
                 val = cast(value.Dict, UP_val)
+                d = NewDict()  # type: Dict[str, value_t]
                 for k, v in iteritems(val.d):
                     d[k] = v
 
                 return value.Dict(d)
 
             elif case(value_e.Obj):
-                d = NewDict()
                 val = cast(Obj, UP_val)
+                d = NewDict()
                 for k, v in iteritems(val.d):
                     d[k] = v
 
                 return value.Dict(d)
 
             elif case(value_e.BashAssoc):
-                d = NewDict()
                 val = cast(value.BashAssoc, UP_val)
+                d = NewDict()
                 for k, s in iteritems(val.d):
                     d[k] = value.Str(s)
 
                 return value.Dict(d)
 
+            elif case(value_e.Frame):
+                val = cast(value.Frame, UP_val)
+                d = NewDict()
+                for k, cell in iteritems(val.frame):
+                    d[k] = cell.val
+
+                return value.Dict(d)
+
         raise error.TypeErr(val, 'dict() expected Dict, Obj, or BashAssoc',
                             rd.BlamePos())
 
@@ -494,52 +544,6 @@ class Glob(vm._Callable):
         return value.List(l)
 
 
-class Shvar_get(vm._Callable):
-    """Look up with dynamic scope."""
-
-    def __init__(self, mem):
-        # type: (state.Mem) -> None
-        vm._Callable.__init__(self)
-        self.mem = mem
-
-    def Call(self, rd):
-        # type: (typed_args.Reader) -> value_t
-        name = rd.PosStr()
-        rd.Done()
-        return state.DynamicGetVar(self.mem, name, scope_e.Dynamic)
-
-
-class GetVar(vm._Callable):
-    """Look up normal scoping rules."""
-
-    def __init__(self, mem):
-        # type: (state.Mem) -> None
-        vm._Callable.__init__(self)
-        self.mem = mem
-
-    def Call(self, rd):
-        # type: (typed_args.Reader) -> value_t
-        name = rd.PosStr()
-        rd.Done()
-        return state.DynamicGetVar(self.mem, name, scope_e.LocalOrGlobal)
-
-
-class EvalExpr(vm._Callable):
-
-    def __init__(self, expr_ev):
-        # type: (expr_eval.ExprEvaluator) -> None
-        self.expr_ev = expr_ev
-
-    def Call(self, rd):
-        # type: (typed_args.Reader) -> value_t
-        lazy = rd.PosExpr()
-        rd.Done()
-
-        result = self.expr_ev.EvalExpr(lazy, rd.LeftParenToken())
-
-        return result
-
-
 class ToJson8(vm._Callable):
 
     def __init__(self, is_j8):
