@@ -5,7 +5,7 @@ User-defined funcs and procs
 from __future__ import print_function
 
 from _devbuild.gen.id_kind_asdl import Id
-from _devbuild.gen.runtime_asdl import cmd_value, ProcArgs
+from _devbuild.gen.runtime_asdl import cmd_value, ProcArgs, Cell
 from _devbuild.gen.syntax_asdl import (proc_sig, proc_sig_e, Param, ParamGroup,
                                        NamedArg, Func, loc, ArgList, expr,
                                        expr_e, expr_t)
@@ -180,7 +180,7 @@ def _EvalNamedArgs(expr_ev, named_exprs):
 def _EvalArgList(
         expr_ev,  # type: expr_eval.ExprEvaluator
         args,  # type: ArgList
-        me=None  # type: Optional[value_t]
+        self_val=None  # type: Optional[value_t]
 ):
     # type: (...) -> Tuple[List[value_t], Optional[Dict[str, value_t]]]
     """Evaluate arg list for funcs.
@@ -195,8 +195,8 @@ def _EvalArgList(
     """
     pos_args = []  # type: List[value_t]
 
-    if me:  # self/this argument
-        pos_args.append(me)
+    if self_val:  # self/this argument
+        pos_args.append(self_val)
 
     _EvalPosArgs(expr_ev, args.pos_args, pos_args)
 
@@ -209,6 +209,8 @@ def _EvalArgList(
 
 def EvalTypedArgsToProc(
         expr_ev,  # type: expr_eval.ExprEvaluator
+        current_frame,  # type: Dict[str, Cell]
+        module_frame,  # type: Dict[str, Cell]
         mutable_opts,  # type: state.MutableOpts
         node,  # type: command.Simple
         proc_args,  # type: ProcArgs
@@ -230,7 +232,8 @@ def EvalTypedArgsToProc(
             # Defer evaluation by wrapping in value.Expr
 
             for exp in ty.pos_args:
-                proc_args.pos_args.append(value.Expr(exp))
+                proc_args.pos_args.append(
+                    value.Expr(exp, current_frame, module_frame))
             # TODO: ...spread is illegal
 
             n1 = ty.named_args
@@ -238,7 +241,8 @@ def EvalTypedArgsToProc(
                 proc_args.named_args = NewDict()
                 for named_arg in n1:
                     name = lexer.TokenVal(named_arg.name)
-                    proc_args.named_args[name] = value.Expr(named_arg.value)
+                    proc_args.named_args[name] = value.Expr(
+                        named_arg.value, current_frame, module_frame)
                 # TODO: ...spread is illegal
 
         else:  # json write (x)
@@ -260,8 +264,9 @@ def EvalTypedArgsToProc(
 
     # p { echo hi } is an unevaluated block
     if node.block:
-        # TODO: conslidate value.Block (holds LiteralBlock) and value.Command
-        proc_args.block_arg = value.Block(node.block)
+        # Attach current frame to command fragment
+        proc_args.block_arg = value.Command(node.block, current_frame,
+                                            module_frame)
 
         # Add location info so the cmd_val looks the same for both:
         #   cd /tmp (; ; ^(echo hi))
@@ -483,16 +488,29 @@ def BindProcArgs(proc, cmd_val, mem):
                 "Proc %r takes no word args, but got %d" %
                 (proc.name, num_word - 1), blame_loc)
 
-    ### Handle typed positional args.  This includes a block arg, if any.
+    ### Handle typed positional args.
 
     if proc_args and proc_args.typed_args:  # blame ( of call site
         blame_loc = proc_args.typed_args.left
 
-    pos_args = proc_args.pos_args if proc_args else None
-    if sig.positional:  # or sig.block_param:
+    if cmd_val.self_obj:
+        pos_args = [cmd_val.self_obj]  # type: List[value_t]
+        if proc_args:
+            pos_args.extend(proc_args.pos_args)
+    else:
+        if proc_args:  # save an allocation in this common case
+            pos_args = proc_args.pos_args
+        else:
+            pos_args = []
+
+    if sig.positional:
         _BindTyped(proc.name, sig.positional, proc.defaults.for_typed,
                    pos_args, mem, blame_loc)
     else:
+        if cmd_val.self_obj is not None:
+            raise error.Expr(
+                "Using proc %r as __invoke__ requires a 'self' param" %
+                proc.name, blame_loc)
         if pos_args is not None:
             num_pos = len(pos_args)
             if num_pos != 0:
