@@ -128,6 +128,15 @@ class history(object):
     self.r = cast(value.Bool, attrs['r']).b  # type: bool
 
 
+class invoke(object):
+  def __init__(self, attrs):
+    # type: (Dict[str, value_t]) -> None
+
+    self.builtin = cast(value.Bool, attrs['builtin']).b  # type: bool
+    self.extern_ = cast(value.Bool, attrs['extern_']).b  # type: bool
+    self.proc_like = cast(value.Bool, attrs['proc_like']).b  # type: bool
+
+
 class jobs(object):
   def __init__(self, attrs):
     # type: (Dict[str, value_t]) -> None
@@ -347,6 +356,13 @@ class unset(object):
     self.v = cast(value.Bool, attrs['v']).b  # type: bool
 
 
+class use(object):
+  def __init__(self, attrs):
+    # type: (Dict[str, value_t]) -> None
+
+    self.extern_ = cast(value.Bool, attrs['extern_']).b  # type: bool
+
+
 class wait(object):
   def __init__(self, attrs):
     # type: (Dict[str, value_t]) -> None
