@@ -13,6 +13,8 @@ class doc_e(object):
   Indent = 3
   Concat = 4
   Group = 5
+  IfFlat = 6
+  Flat = 7
 
 _doc_str = {
   1: 'Break',
@@ -20,6 +22,8 @@ _doc_str = {
   3: 'Indent',
   4: 'Concat',
   5: 'Group',
+  6: 'IfFlat',
+  7: 'Flat',
 }
 
 def doc_str(tag, dot=True):
@@ -288,6 +292,111 @@ class doc(object):
       # type: (Optional[TraversalState]) -> hnode_t
       return self._AbbreviatedTree(trav=trav)
   
+  class IfFlat(doc_t):
+    _type_tag = 6
+    __slots__ = ('flat_mdoc', 'nonflat_mdoc')
+  
+    def __init__(self, flat_mdoc, nonflat_mdoc):
+      # type: (MeasuredDoc, MeasuredDoc) -> None
+      self.flat_mdoc = flat_mdoc
+      self.nonflat_mdoc = nonflat_mdoc
+  
+    @staticmethod
+    def CreateNull(alloc_lists=False):
+      # type: () -> doc.IfFlat
+      return doc.IfFlat(cast(MeasuredDoc, None), cast(MeasuredDoc, None))
+  
+    def PrettyTree(self, trav=None):
+      # type: (Optional[TraversalState]) -> hnode_t
+      trav = trav or TraversalState()
+      heap_id = id(self)
+      if heap_id in trav.seen:
+        return hnode.AlreadySeen(heap_id)
+      trav.seen[heap_id] = True
+      out_node = NewRecord('doc.IfFlat')
+      L = out_node.fields
+  
+      assert self.flat_mdoc is not None
+      x0 = self.flat_mdoc.PrettyTree(trav=trav)
+      L.append(Field('flat_mdoc', x0))
+  
+      assert self.nonflat_mdoc is not None
+      x1 = self.nonflat_mdoc.PrettyTree(trav=trav)
+      L.append(Field('nonflat_mdoc', x1))
+  
+      return out_node
+  
+    def _AbbreviatedTree(self, trav=None):
+      # type: (Optional[TraversalState]) -> hnode_t
+      trav = trav or TraversalState()
+      heap_id = id(self)
+      if heap_id in trav.seen:
+        return hnode.AlreadySeen(heap_id)
+      trav.seen[heap_id] = True
+      out_node = NewRecord('doc.IfFlat')
+      L = out_node.fields
+      assert self.flat_mdoc is not None
+      x0 = self.flat_mdoc.AbbreviatedTree(trav=trav)
+      L.append(Field('flat_mdoc', x0))
+  
+      assert self.nonflat_mdoc is not None
+      x1 = self.nonflat_mdoc.AbbreviatedTree(trav=trav)
+      L.append(Field('nonflat_mdoc', x1))
+  
+      return out_node
+  
+    def AbbreviatedTree(self, trav=None):
+      # type: (Optional[TraversalState]) -> hnode_t
+      return self._AbbreviatedTree(trav=trav)
+  
+  class Flat(doc_t):
+    _type_tag = 7
+    __slots__ = ('mdoc',)
+  
+    def __init__(self, mdoc):
+      # type: (MeasuredDoc) -> None
+      self.mdoc = mdoc
+  
+    @staticmethod
+    def CreateNull(alloc_lists=False):
+      # type: () -> doc.Flat
+      return doc.Flat(cast(MeasuredDoc, None))
+  
+    def PrettyTree(self, trav=None):
+      # type: (Optional[TraversalState]) -> hnode_t
+      trav = trav or TraversalState()
+      heap_id = id(self)
+      if heap_id in trav.seen:
+        return hnode.AlreadySeen(heap_id)
+      trav.seen[heap_id] = True
+      out_node = NewRecord('doc.Flat')
+      L = out_node.fields
+  
+      assert self.mdoc is not None
+      x0 = self.mdoc.PrettyTree(trav=trav)
+      L.append(Field('mdoc', x0))
+  
+      return out_node
+  
+    def _AbbreviatedTree(self, trav=None):
+      # type: (Optional[TraversalState]) -> hnode_t
+      trav = trav or TraversalState()
+      heap_id = id(self)
+      if heap_id in trav.seen:
+        return hnode.AlreadySeen(heap_id)
+      trav.seen[heap_id] = True
+      out_node = NewRecord('doc.Flat')
+      L = out_node.fields
+      assert self.mdoc is not None
+      x0 = self.mdoc.AbbreviatedTree(trav=trav)
+      L.append(Field('mdoc', x0))
+  
+      return out_node
+  
+    def AbbreviatedTree(self, trav=None):
+      # type: (Optional[TraversalState]) -> hnode_t
+      return self._AbbreviatedTree(trav=trav)
+  
   pass
 
 class MeasuredDoc(pybase.CompoundObj):
