@@ -3,8 +3,8 @@ from mycpp import mops
 from typing import Optional, List, Tuple, Dict, Any, cast, TYPE_CHECKING
 
 if TYPE_CHECKING:
-  from _devbuild.gen.syntax_asdl import loc_t, Token, expr_t, word_t, command_t, CompoundWord, DoubleQuoted, ArgList, re_t, redir_loc_t, proc_sig_t, LiteralBlock, Func
-  from _devbuild.gen.value_asdl import value_t
+  from _devbuild.gen.syntax_asdl import loc_t, Token, expr_t, word_t, command_t, CompoundWord, DoubleQuoted, ArgList, re_t, redir_loc_t, proc_sig_t, Func
+  from _devbuild.gen.value_asdl import value_t, Obj
 
 from _devbuild.gen.id_kind_asdl import Id_t
 from _devbuild.gen.id_kind_asdl import Id_str
@@ -40,19 +40,20 @@ class cmd_value_t(pybase.CompoundObj):
 class cmd_value(object):
   class Argv(cmd_value_t):
     _type_tag = 1
-    __slots__ = ('argv', 'arg_locs', 'is_last_cmd', 'proc_args')
+    __slots__ = ('argv', 'arg_locs', 'is_last_cmd', 'self_obj', 'proc_args')
   
-    def __init__(self, argv, arg_locs, is_last_cmd, proc_args):
-      # type: (List[str], List[CompoundWord], bool, Optional[ProcArgs]) -> None
+    def __init__(self, argv, arg_locs, is_last_cmd, self_obj, proc_args):
+      # type: (List[str], List[CompoundWord], bool, Optional[Obj], Optional[ProcArgs]) -> None
       self.argv = argv
       self.arg_locs = arg_locs
       self.is_last_cmd = is_last_cmd
+      self.self_obj = self_obj
       self.proc_args = proc_args
   
     @staticmethod
     def CreateNull(alloc_lists=False):
       # type: () -> cmd_value.Argv
-      return cmd_value.Argv([] if alloc_lists else cast('List[str]', None), [] if alloc_lists else cast('List[CompoundWord]', None), False, cast('Optional[ProcArgs]', None))
+      return cmd_value.Argv([] if alloc_lists else cast('List[str]', None), [] if alloc_lists else cast('List[CompoundWord]', None), False, cast('Optional[Obj]', None), cast('Optional[ProcArgs]', None))
   
     def PrettyTree(self, trav=None):
       # type: (Optional[TraversalState]) -> hnode_t
@@ -81,9 +82,13 @@ class cmd_value(object):
       x2 = hnode.Leaf('T' if self.is_last_cmd else 'F', color_e.OtherConst)
       L.append(Field('is_last_cmd', x2))
   
+      if self.self_obj is not None:  # Optional
+        x3 = self.self_obj.PrettyTree(trav=trav)
+        L.append(Field('self_obj', x3))
+  
       if self.proc_args is not None:  # Optional
-        x3 = self.proc_args.PrettyTree(trav=trav)
-        L.append(Field('proc_args', x3))
+        x4 = self.proc_args.PrettyTree(trav=trav)
+        L.append(Field('proc_args', x4))
   
       return out_node
   
@@ -113,9 +118,13 @@ class cmd_value(object):
       x2 = hnode.Leaf('T' if self.is_last_cmd else 'F', color_e.OtherConst)
       L.append(Field('is_last_cmd', x2))
   
+      if self.self_obj is not None:  # Optional
+        x3 = self.self_obj.AbbreviatedTree(trav=trav)
+        L.append(Field('self_obj', x3))
+  
       if self.proc_args is not None:  # Optional
-        x3 = self.proc_args.AbbreviatedTree(trav=trav)
-        L.append(Field('proc_args', x3))
+        x4 = self.proc_args.AbbreviatedTree(trav=trav)
+        L.append(Field('proc_args', x4))
   
       return out_node
   
