@@ -6,10 +6,9 @@
 #
 #   http://www.apache.org/licenses/LICENSE-2.0
 """
-id_kind_def.py - Id and Kind definitions, used for Token, Word, Nodes, etc.
+id_kind_def.py - Id and Kind definitions, stored in Token
 
-NOTE: If this changes, the lexer may need to be recompiled with
-build/codegen.sh lexer.
+NOTE: If this file changes, rebuild it with build/py.sh all
 """
 from __future__ import print_function
 
@@ -110,10 +109,10 @@ class IdSpec(object):
     ):
         # type: (...) -> None
         """
-    Args:
-      kind_name: string
-      arg_type_pairs: dictionary of bool_arg_type_e -> []
-    """
+        Args:
+          kind_name: string
+          arg_type_pairs: dictionary of bool_arg_type_e -> []
+        """
         lexer_pairs = []
         num_tokens = 0
         for arg_type, pairs in arg_type_pairs:
@@ -233,13 +232,12 @@ def AddKinds(spec):
     #   $'\z'  Such bad codes are accepted when parse_backslash is on
     #          (default in OSH), so we have to lex them.
     #  (x == y) should used === or ~==
-    spec.AddKind('Unknown', ['Tok', 'Backslash', 'DEqual'])
+    spec.AddKind('Unknown', ['Tok', 'Backslash', 'DEqual', 'DAmp', 'DPipe', 'DDot'])
 
     spec.AddKind('Eol', ['Tok'])  # no more tokens on line (\0)
 
-    # HereTabs is for stripping tabs on each line of <<- , while preserving the
-    # "lossless invariant"
-    spec.AddKind('Ignored', ['LineCont', 'Space', 'Comment', 'HereTabs'])
+    # Ignored_Newline is for J8 lexing to count lines
+    spec.AddKind('Ignored', ['LineCont', 'Space', 'Comment', 'Newline'])
 
     # Id.WS_Space is for lex_mode_e.ShCommand; Id.Ignored_Space is for
     # lex_mode_e.Arith
@@ -249,6 +247,7 @@ def AddKinds(spec):
         'Lit',
         [
             'Chars',
+            'CharsWithoutPrefix',  # for stripping leading whitespace
             'VarLike',
             'ArrayLhsOpen',
             'ArrayLhsClose',
@@ -256,8 +255,7 @@ def AddKinds(spec):
             'AtLBracket',  # @[split(x)]
             'AtLBraceDot',  # @{.myproc arg1} should be builtin_sub
             'Other',
-            'EscapedChar',
-            'RegexMeta',
+            'EscapedChar',  # \* is escaped
             'LBracket',
             'RBracket',  # for assoc array literals, static globs
             'Star',
@@ -302,6 +300,8 @@ def AddKinds(spec):
             'DPipe',  # ||
             'Semi',  # ;
             'DSemi',  # ;; for case
+            'SemiAmp',  # ;& for case
+            'DSemiAmp',  # ;;& for case
             'LParen',  # For subshell.  Not Kind.Left because it's NOT a WordPart.
             'RParen',  # Default, will be translated to Id.Right_*
             'DLeftParen',
@@ -333,7 +333,8 @@ def AddKinds(spec):
             'Float',
             'Bang',  # eggex !digit, ![a-z]
             'Dot',
-            'DDot',
+            'DDotLessThan',
+            'DDotEqual',
             'Colon',  # mylist:pop()
             'RArrow',
             'RDArrow',
@@ -372,12 +373,6 @@ def AddKinds(spec):
             'Func',  # For function literals
             'Capture',
             'As',
-
-            # Tea-specific
-            'While',
-            'Break',
-            'Continue',
-            'Return'
         ])
 
     # For C-escaped strings.
@@ -397,13 +392,16 @@ def AddKinds(spec):
             'Unicode8',  # bash
             'UBraced',
             'Pound',  # YSH
-            'Literals',
             'AsciiControl',  # \x01-\x1f, what's disallowed in JSON
         ])
 
-    # Regular expression primtiives.
+    # For lex_mode_e.BashRegex
+    # Bash treats ( | ) as special, and space is allowed within ()
+    # Note Id.Op_RParen -> Id.Right_BashRegex with lexer hint
+    spec.AddKind('BashRegex', ['LParen', 'AllowedInParens'])
+
     spec.AddKind(
-        'Re',
+        'Eggex',
         [
             'Start',  # ^ or %start
             'End',  # $ or %end
@@ -437,6 +435,7 @@ def AddKinds(spec):
         'Left',
         [
             'DoubleQuote',
+            'JDoubleQuote',  # j" for J8 notation
             'SingleQuote',  # ''
             'DollarSingleQuote',  # $'' for \n escapes
             'RSingleQuote',  # r''
@@ -445,6 +444,7 @@ def AddKinds(spec):
 
             # Multiline versions
             'TDoubleQuote',  # """ """
+            'DollarTDoubleQuote',  # $""" """
             'TSingleQuote',  # ''' '''
             'RTSingleQuote',  # r''' '''
             'UTSingleQuote',  # u''' '''
@@ -452,6 +452,7 @@ def AddKinds(spec):
             'Backtick',  # `
             'DollarParen',  # $(
             'DollarBrace',  # ${
+            'DollarBraceZsh',  # ${(foo)
             'DollarDParen',  # $((
             'DollarBracket',  # $[ - synonym for $(( in bash and zsh
             'DollarDoubleQuote',  # $" for bash localized strings
@@ -484,6 +485,7 @@ def AddKinds(spec):
             'CasePat',  # )
             'ShArrayLiteral',  # )
             'ExtGlob',  # )
+            'BashRegexGroup',  # )
             'BlockLiteral',  # } that matches &{ echo hi }
         ])
 
@@ -550,7 +552,7 @@ def AddKinds(spec):
         ])
 
     spec.AddKindPairs(
-        'VOpOil',
+        'VOpYsh',
         [
             ('Pipe', '|'),  # ${x|html}
             ('Space', ' '),  # ${x %.3f}
@@ -620,6 +622,7 @@ def AddKinds(spec):
             # later: Auto?
             'Call',
             'Proc',
+            'Typed',
             'Func',
 
             # builtins, NOT keywords: use, fork, wait, etc.
@@ -689,12 +692,14 @@ def AddKinds(spec):
             'Float',  # Number
 
             # High level tokens for "" b'' u''
-            # We don't distinguish them in the parser, because we parse JSON in
-            # the lexer.
+            # We don't distinguish them in the parser, because we recognize
+            # strings in the lexer.
             'String',
 
             # JSON8 and NIL8
             'Identifier',
+            'Newline',  # J8 Lines only, similar to Op_Newline
+            'Tab',  # Reserved for TSV8
 
             # NIL8 only
             'LParen',
@@ -703,6 +708,8 @@ def AddKinds(spec):
             'Operator',
         ])
 
+    spec.AddKind('ShNumber', ['Dec', 'Hex', 'Oct', 'BaseN'])
+
 
 # Shared between [[ and test/[.
 _UNARY_STR_CHARS = 'zn'  # -z -n
@@ -727,6 +734,15 @@ def AddBoolKinds(spec):
         (bool_arg_type_e.Path, _Dash(list(_UNARY_PATH_CHARS))),
     ])
 
+    Id = spec.id_str2int
+
+    # test --true and test --false have no single letter flags.  They need no
+    # lexing.
+    for long_flag in ('true', 'false'):
+        id_name = 'BoolUnary_%s' % long_flag
+        spec._AddId(id_name)
+        spec.AddBoolOp(Id[id_name], bool_arg_type_e.Str)
+
     spec.AddBoolKind('BoolBinary', [
         (bool_arg_type_e.Str, [
             ('GlobEqual', '='),
@@ -738,7 +754,6 @@ def AddBoolKinds(spec):
         (bool_arg_type_e.Int, _Dash(_BINARY_INT)),
     ])
 
-    Id = spec.id_str2int
     # logical, arity, arg_type
     spec.AddBoolOp(Id['Op_DAmp'], bool_arg_type_e.Undefined)
     spec.AddBoolOp(Id['Op_DPipe'], bool_arg_type_e.Undefined)
