@@ -6,8 +6,8 @@ from __future__ import print_function
 
 from _devbuild.gen import arg_types
 from _devbuild.gen.runtime_asdl import cmd_value
-from _devbuild.gen.syntax_asdl import command_e, BraceGroup, loc
-from _devbuild.gen.value_asdl import value, value_e, value_t
+from _devbuild.gen.syntax_asdl import command_e, BraceGroup
+from _devbuild.gen.value_asdl import value, value_e
 from asdl import format as fmt
 from core import error
 from core.error import e_usage
@@ -19,7 +19,7 @@ from frontend import flag_util
 from frontend import match
 from frontend import typed_args
 from mycpp import mylib
-from mycpp.mylib import tagswitch, log, iteritems
+from mycpp.mylib import log, iteritems
 
 from typing import TYPE_CHECKING, cast
 if TYPE_CHECKING:
@@ -67,27 +67,17 @@ class Pp(_Builtin):
 
         blame_tok = rd.LeftParenToken()
 
-        # It might be nice to add a string too, like
-        # pp 'my annotation' (actual)
-        # But the var name should meaningful in most cases
-
-        UP_val = val
-        result = None  # type: value_t
-        with tagswitch(val) as case:
-            if case(value_e.Expr):  # Destructured assert [true === f()]
-                val = cast(value.Expr, UP_val)
-
-                # In this case, we could get the unevaluated code string and
-                # print it.  Although quoting the line seems enough.
-                result = self.expr_ev.EvalExpr(val.e, blame_tok)
-            else:
-                result = val
-
         # Show it with location
+        # It looks like
+        #   pp (42)
+        #      ^
+        # [ stdin ]:5: (Int)   42
+        # We could also print with ! or -^-
+
         self.stdout_.write('\n')
         excerpt, prefix = ui.CodeExcerptAndPrefix(blame_tok)
         self.stdout_.write(excerpt)
-        ui.PrettyPrintValue(prefix, result, self.stdout_)
+        ui.PrettyPrintValue(prefix, val, self.stdout_)
 
         return 0
 
@@ -100,8 +90,7 @@ class Pp(_Builtin):
         action, action_loc = arg_r.Peek2()
 
         # Special cases
-        # pp (x) quotes its code location
-        # pp [x] also evaluates
+        # pp (x) quotes its code location, can also be pp [x]
         if action is None:
             return self._PrettyPrint(cmd_val)
 
@@ -190,6 +179,11 @@ class Pp(_Builtin):
 
             return 0
 
+        if action == 'frame_vars_':  # Print names in current frame, for testing
+            top = self.mem.var_stack[-1]
+            print('    [frame_vars_] %s' % ' '.join(top.keys()))
+            return 0
+
         if action == 'gc-stats_':
             print('TODO')
             return 0
@@ -198,20 +192,23 @@ class Pp(_Builtin):
             names, locs = arg_r.Rest2()
             if len(names):
                 for i, name in enumerate(names):
-                    node = self.procs.Get(name)
+                    node, _ = self.procs.GetInvokable(name)
                     if node is None:
                         self.errfmt.Print_('Invalid proc %r' % name,
                                            blame_loc=locs[i])
                         return 1
             else:
-                names = self.procs.GetNames()
+                names = self.procs.InvokableNames()
 
             # TSV8 header
             print('proc_name\tdoc_comment')
             for name in names:
-                proc = self.procs.Get(name)  # must exist
-                #log('Proc %s', proc)
-                body = proc.body
+                proc_val, _ = self.procs.GetInvokable(name)  # must exist
+                if proc_val.tag() != value_e.Proc:
+                    continue  # can't be value.BuiltinProc
+                user_proc = cast(value.Proc, proc_val)
+
+                body = user_proc.body
 
                 # TODO: not just command.ShFunction, but command.Proc!
                 doc = ''
@@ -281,14 +278,11 @@ class Write(_Builtin):
         return 0
 
 
-class Fopen(vm._Builtin):
-    """fopen does nothing but run a block.
+class RunBlock(vm._Builtin):
+    """Used for 'redir' builtin
 
     It's used solely for its redirects.
-        fopen >out.txt { echo hi }
-
-    It's a subset of eval
-        eval >out.txt { echo hi }
+        redir >out.txt { echo hi }
     """
 
     def __init__(self, mem, cmd_ev):
@@ -298,13 +292,10 @@ class Fopen(vm._Builtin):
 
     def Run(self, cmd_val):
         # type: (cmd_value.Argv) -> int
-        _, arg_r = flag_util.ParseCmdVal('fopen',
+        _, arg_r = flag_util.ParseCmdVal('redir',
                                          cmd_val,
                                          accept_typed_args=True)
 
-        cmd = typed_args.OptionalBlock(cmd_val)
-        if not cmd:
-            raise error.Usage('expected a block', loc.Missing)
-
-        unused = self.cmd_ev.EvalCommand(cmd)
+        cmd_frag = typed_args.RequiredBlockAsFrag(cmd_val)
+        unused = self.cmd_ev.EvalCommandFrag(cmd_frag)
         return 0
