@@ -109,7 +109,7 @@ class Json(vm._Builtin):
                 #log('VAR %s', var_name)
                 blame_loc = cmd_val.arg_locs[0]
                 place = value.Place(LeftName(var_name, blame_loc),
-                                    self.mem.TopNamespace())
+                                    self.mem.CurrentFrame())
 
             if not arg_r.AtEnd():
                 e_usage('read got too many args', arg_r.Location())
