@@ -679,6 +679,17 @@ JSON_STR_DEF = [
     R(r'[^\\"\x00-\x1F]+', Id.Lit_Chars),
 ]
 
+_WHITESPACE = r'[ \t\r\n]*'  # ASCII whitespace doesn't have legacy \f \v
+
+SH_NUMBER_DEF = [
+    R('0', Id.ShNumber_Dec),  # not octal
+    R(r'[1-9][0-9]*', Id.ShNumber_Dec),
+    R(r'0[0-7]+', Id.ShNumber_Oct),
+    R(r'0x[0-9A-Fa-f]+', Id.ShNumber_Hex),
+    R(r'[1-9][0-9]*#[0-9a-zA-Z@_]+', Id.ShNumber_BaseN),
+    R(r'[^\0]', Id.Unknown_Tok),  # any other char
+]
+
 OCTAL3_RE = r'\\[0-7]{1,3}'
 
 # https://www.gnu.org/software/bash/manual/html_node/Controlling-the-PromptEvaluator.html#Controlling-the-PromptEvaluator
@@ -953,25 +964,29 @@ _EXPR_NEWLINE_COMMENT = [
     R(r'[ \t\r]+', Id.Ignored_Space),
 ]
 
-_WHITESPACE = r'[ \t\r\n]*'  # ASCII whitespace doesn't have legacy \f \v
+# Note: if you call match.LooksLikeInteger(s), mops.FromStr(s) may still
+# fail.  However you should call BOTH, because we don't rely want to rely on
+# the underlying stroll() to define the language accepted.
+LOOKS_LIKE_INTEGER = _WHITESPACE + '-?[0-9]+' + _WHITESPACE
 
+# TODO: use for YSH comparison operators > >= < <=
+#
 # Python allows 0 to be written 00 or 0_0_0, which is weird.  But let's be
 # consistent, and avoid '00' turning into a float!
-_DECIMAL_INT_RE = r'[0-9](_?[0-9])*'
+_YSH_DECIMAL_INT_RE = r'[0-9](_?[0-9])*'
 
-# Used for YSH comparison operators > >= < <=
-LOOKS_LIKE_INTEGER = _WHITESPACE + '-?' + _DECIMAL_INT_RE + _WHITESPACE
+LOOKS_LIKE_YSH_INT = _WHITESPACE + '-?' + _YSH_DECIMAL_INT_RE + _WHITESPACE
 
-_FLOAT_RE = (
-    _DECIMAL_INT_RE +
+_YSH_FLOAT_RE = (
+    _YSH_DECIMAL_INT_RE +
     # Unlike Python, exponent can't be like 42e5_000.  There's no use because
     # 1e309 is already inf.  Let's keep our code simple.
-    r'(\.' + _DECIMAL_INT_RE + ')?([eE][+\-]?[0-9]+)?')
+    r'(\.' + _YSH_DECIMAL_INT_RE + ')?([eE][+\-]?[0-9]+)?')
 
-# Ditto, used for comparison operators
+# Ditto, used for YSH comparison operators
 # Added optional Optional -?
 # Example: -3_000_000.000_001e12
-LOOKS_LIKE_FLOAT = _WHITESPACE + '-?' + _FLOAT_RE + _WHITESPACE
+LOOKS_LIKE_YSH_FLOAT = _WHITESPACE + '-?' + _YSH_FLOAT_RE + _WHITESPACE
 
 # Python 3 float literals:
 
@@ -1000,13 +1015,13 @@ LEXER_DEF[lex_mode_e.Expr] = \
     # octdigit     ::=  "0"..."7"
     # hexdigit     ::=  digit | "a"..."f" | "A"..."F"
 
-    R(_DECIMAL_INT_RE, Id.Expr_DecInt),
+    R(_YSH_DECIMAL_INT_RE, Id.Expr_DecInt),
 
     R(r'0[bB](_?[01])+', Id.Expr_BinInt),
     R(r'0[oO](_?[0-7])+', Id.Expr_OctInt),
     R(r'0[xX](_?[0-9a-fA-F])+', Id.Expr_HexInt),
 
-    R(_FLOAT_RE, Id.Expr_Float),
+    R(_YSH_FLOAT_RE, Id.Expr_Float),
 
     # These can be looked up as keywords separately, so you enforce that they have
     # space around them?
@@ -1042,6 +1057,9 @@ LEXER_DEF[lex_mode_e.Expr] = \
 
     C('==', Id.Unknown_DEqual),  # user must choose === or ~==
 
+    C('&&', Id.Unknown_DAmp),
+    C('||', Id.Unknown_DPipe),
+
     # Bitwise operators
     C('&', Id.Arith_Amp),
     C('|', Id.Arith_Pipe),
@@ -1086,10 +1104,12 @@ LEXER_DEF[lex_mode_e.Expr] = \
     C('//', Id.Expr_DSlash),  # For YSH integer division
     C('~==', Id.Expr_TildeDEqual),  # approximate equality
 
-    C('.', Id.Expr_Dot),      # d.key is alias for d['key']
-    C('..', Id.Expr_DDot),    # range 1..5
-    C('->', Id.Expr_RArrow),  # s->startswith()
-    C('$', Id.Expr_Dollar),   # legacy regex end: /d+ $/ (better written /d+ >/
+    C('.', Id.Expr_Dot),             # d.key is alias for d['key']
+    C('..', Id.Unknown_DDot),        # legacy half-open range 1..5
+    C('..<', Id.Expr_DDotLessThan),  # half-open range 1..<5
+    C('..=', Id.Expr_DDotEqual),     # closed range 1..5
+    C('->', Id.Expr_RArrow),         # s->startswith()
+    C('$', Id.Expr_Dollar),          # legacy regex end: /d+ $/ (better written /d+ >/
 
     # Reserved this.  Go uses it for channels, etc.
     # I guess it conflicts with -4<-3, but that's OK -- spaces suffices.
