@@ -27,7 +27,7 @@ from _devbuild.gen.syntax_asdl import (
     word_part_t,
     CompoundWord,
     Token,
-    NameTok,
+    SimpleVarSub,
     ShArrayLiteral,
     SingleQuoted,
     DoubleQuoted,
@@ -121,6 +121,11 @@ def TokenForCommand(node):
     UP_node = node  # type: command_t
     tag = node.tag()
 
+    if tag == command_e.Redirect:
+        node = cast(command.Redirect, UP_node)
+        first = node.redirects[0]
+        return first.op
+
     if tag == command_e.Sentence:
         node = cast(command.Sentence, UP_node)
         #log("node.child %s", node.child)
@@ -181,9 +186,9 @@ def TokenForArith(node):
     UP_node = node
     with tagswitch(node) as case:
         if case(arith_expr_e.VarSub):
-            vsub = cast(NameTok, UP_node)
+            vsub = cast(Token, UP_node)
             # $(( x ))
-            return vsub.left
+            return vsub
 
         elif case(arith_expr_e.Word):
             w = cast(CompoundWord, UP_node)
@@ -195,10 +200,7 @@ def TokenForArith(node):
 
         elif case(arith_expr_e.Binary):
             node = cast(arith_expr.Binary, UP_node)
-
-            # TODO: should blame op
-            # blaming left is arbitrary, but better than nothing
-            return TokenForArith(node.left)
+            return TokenForArith(node.op)
 
         elif case(arith_expr_e.TernaryOp):
             node = cast(arith_expr.TernaryOp, UP_node)
@@ -239,8 +241,8 @@ def LeftTokenForWordPart(part):
             return part.left
 
         elif case(word_part_e.SimpleVarSub):
-            part = cast(NameTok, UP_part)
-            return part.left
+            part = cast(SimpleVarSub, UP_part)
+            return part.tok
 
         elif case(word_part_e.BracedVarSub):
             part = cast(BracedVarSub, UP_part)
@@ -313,9 +315,9 @@ def _RightTokenForWordPart(part):
             return part.right  # right "
 
         elif case(word_part_e.SimpleVarSub):
-            part = cast(NameTok, UP_part)
+            part = cast(SimpleVarSub, UP_part)
             # left and right are the same for $myvar
-            return part.left
+            return part.tok
 
         elif case(word_part_e.BracedVarSub):
             part = cast(BracedVarSub, UP_part)
@@ -477,6 +479,10 @@ def TokenForExpr(node):
             node = cast(expr.Var, UP_node)
             return node.left
 
+        elif case(expr_e.Place):
+            node = cast(expr.Place, UP_node)
+            return node.blame_tok
+
         elif case(expr_e.CommandSub):
             node = cast(CommandSub, UP_node)
             return node.left_token
@@ -498,8 +504,8 @@ def TokenForExpr(node):
             return node.left
 
         elif case(expr_e.SimpleVarSub):
-            node = cast(NameTok, UP_node)
-            return node.left
+            node = cast(SimpleVarSub, UP_node)
+            return node.tok
 
         elif case(expr_e.Unary):
             node = cast(expr.Unary, UP_node)
@@ -534,8 +540,6 @@ def TokenForExpr(node):
             node = cast(expr.Tuple, UP_node)
             return node.left
 
-            return node.left
-
         elif case(expr_e.Dict):
             node = cast(expr.Dict, UP_node)
             return node.left
