@@ -23,6 +23,7 @@ from core import pyos
 from core import pyutil
 from core import vm
 from frontend import flag_util
+from frontend import match
 from frontend import typed_args
 from mycpp import mops
 from mycpp import mylib
@@ -33,7 +34,8 @@ import posix_ as posix
 from typing import TYPE_CHECKING, List, Tuple, Optional, cast
 if TYPE_CHECKING:
     from core.process import Waiter, ExternalProgram, FdState
-    from core.state import Mem, SearchPath
+    from core import executor
+    from core import state
     from display import ui
 
 _ = log
@@ -158,11 +160,8 @@ class Fork(vm._Builtin):
         if arg is not None:
             e_usage('got unexpected argument %r' % arg, location)
 
-        cmd = typed_args.OptionalBlock(cmd_val)
-        if cmd is None:
-            e_usage('expected a block', loc.Missing)
-
-        return self.shell_ex.RunBackgroundJob(cmd)
+        cmd_frag = typed_args.RequiredBlockAsFrag(cmd_val)
+        return self.shell_ex.RunBackgroundJob(cmd_frag)
 
 
 class ForkWait(vm._Builtin):
@@ -180,17 +179,14 @@ class ForkWait(vm._Builtin):
         if arg is not None:
             e_usage('got unexpected argument %r' % arg, location)
 
-        cmd = typed_args.OptionalBlock(cmd_val)
-        if cmd is None:
-            e_usage('expected a block', loc.Missing)
-
-        return self.shell_ex.RunSubshell(cmd)
+        cmd_frag = typed_args.RequiredBlockAsFrag(cmd_val)
+        return self.shell_ex.RunSubshell(cmd_frag)
 
 
 class Exec(vm._Builtin):
 
     def __init__(self, mem, ext_prog, fd_state, search_path, errfmt):
-        # type: (Mem, ExternalProgram, FdState, SearchPath, ui.ErrorFormatter) -> None
+        # type: (state.Mem, ExternalProgram, FdState, executor.SearchPath, ui.ErrorFormatter) -> None
         self.mem = mem
         self.ext_prog = ext_prog
         self.fd_state = fd_state
@@ -206,7 +202,11 @@ class Exec(vm._Builtin):
             self.fd_state.MakePermanent()
             return 0
 
-        environ = self.mem.GetExported()
+        environ = self.mem.GetEnv()
+        if 0:
+            log('E %r', environ)
+            log('E %r', environ)
+            log('ZZ %r', environ.get('ZZ'))
         i = arg_r.i
         cmd = cmd_val.argv[i]
         argv0_path = self.search_path.CachedLookup(cmd)
@@ -215,7 +215,7 @@ class Exec(vm._Builtin):
 
         # shift off 'exec', and remove typed args because they don't apply
         c2 = cmd_value.Argv(cmd_val.argv[i:], cmd_val.arg_locs[i:],
-                            cmd_val.is_last_cmd, None)
+                            cmd_val.is_last_cmd, cmd_val.self_obj, None)
 
         self.ext_prog.Exec(argv0_path, c2, environ)  # NEVER RETURNS
         # makes mypy and C++ compiler happy
@@ -242,7 +242,7 @@ class Wait(vm._Builtin):
     """
 
     def __init__(self, waiter, job_list, mem, tracer, errfmt):
-        # type: (Waiter, process.JobList, Mem, dev.Tracer, ui.ErrorFormatter) -> None
+        # type: (Waiter, process.JobList, state.Mem, dev.Tracer, ui.ErrorFormatter) -> None
         self.waiter = waiter
         self.job_list = job_list
         self.mem = mem
@@ -524,9 +524,11 @@ class Ulimit(vm._Builtin):
             # In C, RLIM_INFINITY is rlim_t
             limit = mops.FromC(RLIM_INFINITY)
         else:
-            try:
-                big_int = mops.FromStr(s)
-            except ValueError as e:
+            if match.LooksLikeInteger(s):
+                ok, big_int = mops.FromStr2(s)
+                if not ok:
+                    raise error.Usage('Integer too big: %s' % s, s_loc)
+            else:
                 raise error.Usage(
                     "expected a number or 'unlimited', got %r" % s, s_loc)
 
