@@ -17,6 +17,7 @@ from mycpp.mylib import tagswitch, NewDict
 from typing import TYPE_CHECKING, cast, Any, Dict, List
 
 if TYPE_CHECKING:
+    from core import executor
     from display import ui
     from osh.cmd_eval import CommandEvaluator
 
@@ -24,7 +25,7 @@ if TYPE_CHECKING:
 class Shvar(vm._Builtin):
 
     def __init__(self, mem, search_path, cmd_ev):
-        # type: (state.Mem, state.SearchPath, CommandEvaluator) -> None
+        # type: (state.Mem, executor.SearchPath, CommandEvaluator) -> None
         self.mem = mem
         self.search_path = search_path  # to clear PATH
         self.cmd_ev = cmd_ev  # To run blocks
@@ -35,11 +36,9 @@ class Shvar(vm._Builtin):
                                          cmd_val,
                                          accept_typed_args=True)
 
-        cmd = typed_args.OptionalBlock(cmd_val)
-        if not cmd:
-            # TODO: I think shvar LANG=C should just mutate
-            # But should there be a whitelist?
-            raise error.Usage('expected a block', loc.Missing)
+        # TODO: I think shvar LANG=C should just mutate
+        # But should there be a whitelist?
+        cmd_frag = typed_args.RequiredBlockAsFrag(cmd_val)
 
         vars = NewDict()  # type: Dict[str, value_t]
         args, arg_locs = arg_r.Rest2()
@@ -58,7 +57,7 @@ class Shvar(vm._Builtin):
                 self.search_path.ClearCache()
 
         with state.ctx_Eval(self.mem, None, None, vars):
-            unused = self.cmd_ev.EvalCommand(cmd)
+            unused = self.cmd_ev.EvalCommandFrag(cmd_frag)
 
         return 0
 
@@ -99,7 +98,7 @@ class Ctx(vm._Builtin):
     def _Push(self, context, block):
         # type: (Dict[str, value_t], command_t) -> int
         with ctx_Context(self.mem, context):
-            return self.cmd_ev.EvalCommand(block)
+            return self.cmd_ev.EvalCommandFrag(block)
 
     def _Set(self, updates):
         # type: (Dict[str, value_t]) -> int
@@ -137,7 +136,7 @@ class Ctx(vm._Builtin):
 
         if verb == "push":
             context = rd.PosDict()
-            block = rd.RequiredBlock()
+            block = rd.RequiredBlockAsFrag()
             rd.Done()
             arg_r.AtEnd()
 
@@ -176,12 +175,10 @@ class PushRegisters(vm._Builtin):
                                          cmd_val,
                                          accept_typed_args=True)
 
-        cmd = typed_args.OptionalBlock(cmd_val)
-        if not cmd:
-            raise error.Usage('expected a block', loc.Missing)
+        cmd_frag = typed_args.RequiredBlockAsFrag(cmd_val)
 
         with state.ctx_Registers(self.mem):
-            unused = self.cmd_ev.EvalCommand(cmd)
+            unused = self.cmd_ev.EvalCommandFrag(cmd_frag)
 
         # make it "SILENT" in terms of not mutating $?
         # TODO: Revisit this.  It might be better to provide the headless shell
@@ -206,9 +203,9 @@ class Append(vm._Builtin):
         # type: (cmd_value.Argv) -> int
 
         # This means we ignore -- , which is consistent
-        arg, arg_r = flag_util.ParseCmdVal('append',
-                                           cmd_val,
-                                           accept_typed_args=True)
+        _, arg_r = flag_util.ParseCmdVal('append',
+                                         cmd_val,
+                                         accept_typed_args=True)
 
         rd = typed_args.ReaderForProc(cmd_val)
         val = rd.PosValue()
