@@ -17,15 +17,15 @@ from typing import Optional, TYPE_CHECKING
 if TYPE_CHECKING:
     from _devbuild.gen.runtime_asdl import cmd_value
     from frontend.py_readline import Readline
-    from core.ui import ErrorFormatter
-    from core import shell
+    from core import sh_init
+    from display import ui
 
 
 class Bind(vm._Builtin):
     """For :, true, false."""
 
     def __init__(self, readline, errfmt):
-        # type: (Optional[Readline], ErrorFormatter) -> None
+        # type: (Optional[Readline], ui.ErrorFormatter) -> None
         self.readline = readline
         self.errfmt = errfmt
 
@@ -42,8 +42,8 @@ class History(vm._Builtin):
     def __init__(
             self,
             readline,  # type: Optional[Readline]
-            sh_files,  # type: shell.ShellFiles
-            errfmt,  # type: ErrorFormatter
+            sh_files,  # type: sh_init.ShellFiles
+            errfmt,  # type: ui.ErrorFormatter
             f,  # type: mylib.Writer
     ):
         # type: (...) -> None
@@ -59,7 +59,7 @@ class History(vm._Builtin):
         # zsh -c 'history' produces an error.
         readline = self.readline
         if not readline:
-            e_usage("is disabled because Oil wasn't compiled with 'readline'",
+            e_usage("is disabled because Oils wasn't compiled with 'readline'",
                     loc.Missing)
 
         attrs, arg_r = flag_util.ParseCmdVal('history', cmd_val)
