@@ -1,21 +1,25 @@
-#!/usr/bin/env python2
-"""
-val_ops.py
-"""
 from __future__ import print_function
 
+from errno import EINTR
+
 from _devbuild.gen.syntax_asdl import loc, loc_t, command_t
 from _devbuild.gen.value_asdl import (value, value_e, value_t, eggex_ops,
-                                      eggex_ops_t, regex_match, RegexMatch)
+                                      eggex_ops_t, regex_match, RegexMatch,
+                                      Obj)
 from core import error
-from core import ui
+from core.error import e_die
+from display import ui
 from mycpp import mops
-from mycpp.mylib import tagswitch
+from mycpp import mylib
+from mycpp.mylib import tagswitch, log
 from ysh import regex_translate
 
 from typing import TYPE_CHECKING, cast, Dict, List, Optional
 
 import libc
+import posix_ as posix
+
+_ = log
 
 if TYPE_CHECKING:
     from core import state
@@ -71,24 +75,33 @@ def ToDict(val, msg, blame_loc):
     raise error.TypeErr(val, msg, blame_loc)
 
 
-def ToCommand(val, msg, blame_loc):
+def ToCommandFrag(val, msg, blame_loc):
     # type: (value_t, str, loc_t) -> command_t
     UP_val = val
-    if val.tag() == value_e.Command:
-        val = cast(value.Command, UP_val)
+    if val.tag() == value_e.CommandFrag:
+        val = cast(value.CommandFrag, UP_val)
         return val.c
 
     raise error.TypeErr(val, msg, blame_loc)
 
 
-def Stringify(val, blame_loc, prefix=''):
+def Stringify(val, blame_loc, op_desc):
     # type: (value_t, loc_t, str) -> str
     """
-    Used by
+    Args:
+      op_desc: could be empty string ''
+               or 'Expr Sub ' or 'Expr Splice ', with trailing space
 
-    $[x]    stringify operator
-    @[x]    expression splice - each element is stringified
-    @x      splice value
+    Used by:
+
+      $[x]    Expr Sub - stringify operator
+      @[x]    Expr splice - each element is stringified
+      @x      Splice value
+
+      str()         Builtin function
+      join()        Each element is stringified, e.g. join([1,2])
+                    Not sure I like join([null, true]), but it's consistent
+      Str.replace() ^"x = $x" after eggex conversion function
     """
     if blame_loc is None:
         blame_loc = loc.Missing
@@ -113,24 +126,24 @@ def Stringify(val, blame_loc, prefix=''):
 
         elif case(value_e.Float):
             val = cast(value.Float, UP_val)
-            # TODO: what precision does this have?
-            # The default could be like awk or Python, and then we also allow
-            # ${myfloat %.3f} and more.
-            # Python 3 seems to give a few more digits than Python 2 for str(1.0/3)
             s = str(val.f)
 
         elif case(value_e.Eggex):
             val = cast(value.Eggex, UP_val)
             s = regex_translate.AsPosixEre(val)  # lazily converts to ERE
 
-        elif case(value_e.List):
-            raise error.TypeErrVerbose(
-                "%sgot a List, which can't be stringified. Perhaps use @ instead of $, or use join()"
-                % prefix, blame_loc)
-
         else:
+            pass  # mycpp workaround
+
+            if val.tag() == value_e.List:
+                # Special error message for using the wrong sigil, or maybe join
+                raise error.TypeErrVerbose(
+                    "%sgot a List, which can't be stringified (OILS-ERR-203)" %
+                    op_desc, blame_loc)
+
             raise error.TypeErr(
-                val, "%sexpected Null, Bool, Int, Float, Eggex" % prefix,
+                val,
+                "%sexpected one of (Null Bool Int Float Str Eggex)" % op_desc,
                 blame_loc)
 
     return s
@@ -155,7 +168,7 @@ def ToShellArray(val, blame_loc, prefix=''):
             # Note: it would be nice to add the index to the error message
             # prefix, WITHOUT allocating a string for every item
             for item in val.items:
-                strs.append(Stringify(item, blame_loc, prefix=prefix))
+                strs.append(Stringify(item, blame_loc, prefix))
 
         # I thought about getting rid of this to keep OSH and YSH separate,
         # but:
@@ -171,7 +184,7 @@ def ToShellArray(val, blame_loc, prefix=''):
     return strs
 
 
-class _ContainerIter(object):
+class Iterator(object):
     """Interface for various types of for loop."""
 
     def __init__(self):
@@ -186,13 +199,12 @@ class _ContainerIter(object):
         # type: () -> None
         self.i += 1
 
-    def Done(self):
-        # type: () -> int
-        raise NotImplementedError()
-
     def FirstValue(self):
-        # type: () -> value_t
-        """Return Dict key or List value"""
+        # type: () -> Optional[value_t]
+        """Return a value, or None if done
+
+        e.g. return Dict key or List value
+        """
         raise NotImplementedError()
 
     def SecondValue(self):
@@ -201,67 +213,104 @@ class _ContainerIter(object):
         raise AssertionError("Shouldn't have called this")
 
 
-class ArrayIter(_ContainerIter):
+class StdinIterator(Iterator):
+    """ for x in <> { """
+
+    def __init__(self, blame_loc):
+        # type: (loc_t) -> None
+        Iterator.__init__(self)
+        self.blame_loc = blame_loc
+        self.f = mylib.Stdin()
+
+    def FirstValue(self):
+        # type: () -> Optional[value_t]
+
+        # line, eof = read_osh.ReadLineSlowly(None, with_eol=False)
+        try:
+            line = self.f.readline()
+        except (IOError, OSError) as e:  # signals
+            if e.errno == EINTR:
+                # Caller will can run traps with cmd_ev, like ReadLineSlowly
+                return value.Interrupted
+            else:
+                # For possible errors from f.readline(), see
+                #   man read
+                #   man getline
+                # e.g.
+                # - ENOMEM getline() allocation failure
+                # - EISDIR getline() read from directory descriptor!
+                #
+                # Note: the read builtin returns status 1 for EISDIR.
+                #
+                # We'll raise a top-level error like Python.  (Awk prints a
+                # warning message)
+                e_die("I/O error in for <> loop: %s" % posix.strerror(e.errno),
+                      self.blame_loc)
+
+        if len(line) == 0:
+            return None  # Done
+        elif line.endswith('\n'):
+            # TODO: optimize this to prevent extra garbage
+            line = line[:-1]
+
+        return value.Str(line)
+
+
+class ArrayIter(Iterator):
     """ for x in 1 2 3 { """
 
     def __init__(self, strs):
         # type: (List[str]) -> None
-        _ContainerIter.__init__(self)
+        Iterator.__init__(self)
         self.strs = strs
         self.n = len(strs)
 
-    def Done(self):
-        # type: () -> int
-        return self.i == self.n
-
     def FirstValue(self):
-        # type: () -> value_t
+        # type: () -> Optional[value_t]
+        if self.i == self.n:
+            return None
         return value.Str(self.strs[self.i])
 
 
-class RangeIterator(_ContainerIter):
+class RangeIterator(Iterator):
     """ for x in (m:n) { """
 
     def __init__(self, val):
         # type: (value.Range) -> None
-        _ContainerIter.__init__(self)
+        Iterator.__init__(self)
         self.val = val
 
-    def Done(self):
-        # type: () -> int
-        return self.val.lower + self.i >= self.val.upper
-
     def FirstValue(self):
-        # type: () -> value_t
+        # type: () -> Optional[value_t]
+        if self.val.lower + self.i >= self.val.upper:
+            return None
 
         # TODO: range should be BigInt too
         return value.Int(mops.IntWiden(self.val.lower + self.i))
 
 
-class ListIterator(_ContainerIter):
+class ListIterator(Iterator):
     """ for x in (mylist) { """
 
     def __init__(self, val):
         # type: (value.List) -> None
-        _ContainerIter.__init__(self)
+        Iterator.__init__(self)
         self.val = val
         self.n = len(val.items)
 
-    def Done(self):
-        # type: () -> int
-        return self.i == self.n
-
     def FirstValue(self):
-        # type: () -> value_t
+        # type: () -> Optional[value_t]
+        if self.i == self.n:
+            return None
         return self.val.items[self.i]
 
 
-class DictIterator(_ContainerIter):
+class DictIterator(Iterator):
     """ for x in (mydict) { """
 
     def __init__(self, val):
         # type: (value.Dict) -> None
-        _ContainerIter.__init__(self)
+        Iterator.__init__(self)
 
         # TODO: Don't materialize these Lists
         self.keys = val.d.keys()  # type: List[str]
@@ -270,12 +319,10 @@ class DictIterator(_ContainerIter):
         self.n = len(val.d)
         assert self.n == len(self.keys)
 
-    def Done(self):
-        # type: () -> int
-        return self.i == self.n
-
     def FirstValue(self):
         # type: () -> value_t
+        if self.i == self.n:
+            return None
         return value.Str(self.keys[self.i])
 
     def SecondValue(self):
@@ -336,6 +383,11 @@ def ToBool(val):
 
 def ExactlyEqual(left, right, blame_loc):
     # type: (value_t, value_t, loc_t) -> bool
+
+    if left.tag() == value_e.Float or right.tag() == value_e.Float:
+        raise error.TypeErrVerbose(
+            "Equality isn't defined on Float values (OILS-ERR-202)", blame_loc)
+
     if left.tag() != right.tag():
         return False
 
@@ -359,10 +411,7 @@ def ExactlyEqual(left, right, blame_loc):
             return mops.Equal(left.i, right.i)
 
         elif case(value_e.Float):
-            # Note: could provide floatEquals(), and suggest it
-            # Suggested idiom is abs(f1 - f2) < 0.1
-            raise error.TypeErrVerbose("Equality isn't defined on Float",
-                                       blame_loc)
+            raise AssertionError()
 
         elif case(value_e.Str):
             left = cast(value.Str, UP_left)
@@ -491,4 +540,21 @@ def MatchRegex(left, right, mem):
         return False
 
 
+def IndexMetaMethod(obj):
+    # type: (Obj) -> Optional[value_t]
+    """
+    Returns value.{BuiltinFunc,Func} -- but not callable Obj?
+    """
+    if not obj.prototype:
+        return None
+    index_val = obj.prototype.d.get('__index__')
+    if not index_val:
+        return None
+
+    if index_val.tag() not in (value_e.BuiltinFunc, value_e.Func):
+        return None
+
+    return index_val
+
+
 # vim: sw=4
