@@ -151,7 +151,7 @@ def _DetectMetaBuiltinStr(s):
 
     Fundamentally, assignment builtins have different WORD EVALUATION RULES
     for a=$x (no word splitting), so it seems hard to do this in
-    meta_osh.Builtin() or meta_osh.Command()
+    meta_oils.Builtin() or meta_oils.Command()
     """
     return (consts.LookupNormalBuiltin(s)
             in (builtin_i.builtin, builtin_i.command))
@@ -234,10 +234,10 @@ def _ValueToPartValue(val, quoted, part_loc):
             return part_value.Array(val.d.values())
 
         # Cases added for YSH
-        # value_e.List is also here - we use val_ops.stringify()s err message
+        # value_e.List is also here - we use val_ops.Stringify()s err message
         elif case(value_e.Null, value_e.Bool, value_e.Int, value_e.Float,
                   value_e.Eggex, value_e.List):
-            s = val_ops.Stringify(val, loc.Missing)
+            s = val_ops.Stringify(val, loc.Missing, 'Word eval ')
             return Piece(s, quoted, not quoted)
 
         else:
@@ -474,12 +474,12 @@ class TildeEvaluator(object):
         Important: the libc call can FAIL, which is why we prefer $HOME.  See issue
         #1578.
         """
-        # First look up the HOME var, then ask the OS.  This is what bash does.
-        val = self.mem.GetValue('HOME')
-        UP_val = val
-        if val.tag() == value_e.Str:
-            val = cast(value.Str, UP_val)
-            return val.s
+        # First look up the HOME var, ENV.HOME, ...
+        s = self.mem.env_config.Get('HOME')
+        if s is not None:
+            return s
+
+        # Then ask the OS.  This is what bash does.
         return pyos.GetMyHomeDir()
 
     def Eval(self, part):
@@ -1998,6 +1998,13 @@ class AbstractWordEvaluator(StringWordEvaluator):
 
         will_glob = not self.exec_opts.noglob()
 
+        if 0:
+            log('---')
+            log('FRAME')
+            for i, piece in enumerate(frame):
+                log('(%d) %s', i, piece)
+            log('')
+
         # Array of strings, some of which are BOTH IFS-escaped and GLOB escaped!
         frags = []  # type: List[str]
         for piece in frame:
@@ -2016,6 +2023,13 @@ class AbstractWordEvaluator(StringWordEvaluator):
 
             frags.append(frag)
 
+        if 0:
+            log('---')
+            log('FRAGS')
+            for i, frag in enumerate(frags):
+                log('(%d) %s', i, frag)
+            log('')
+
         flat = ''.join(frags)
         #log('flat: %r', flat)
 
@@ -2222,7 +2236,7 @@ class AbstractWordEvaluator(StringWordEvaluator):
                     strs.append(''.join(tmp))  # no split or glob
                     locs.append(w)
 
-        return cmd_value.Argv(strs, locs, is_last_cmd, None)
+        return cmd_value.Argv(strs, locs, is_last_cmd, None, None)
 
     def EvalWordSequence2(self, words, is_last_cmd, allow_assign=False):
         # type: (List[CompoundWord], bool, bool) -> cmd_value_t
@@ -2326,7 +2340,7 @@ class AbstractWordEvaluator(StringWordEvaluator):
         # A non-assignment command.
         # NOTE: Can't look up builtins here like we did for assignment, because
         # functions can override builtins.
-        return cmd_value.Argv(strs, locs, is_last_cmd, None)
+        return cmd_value.Argv(strs, locs, is_last_cmd, None, None)
 
     def EvalWordSequence(self, words):
         # type: (List[CompoundWord]) -> List[str]
